# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING


class Module:
    if TYPE_CHECKING:
        import testinfra.host

        _host: testinfra.host.Host

    @classmethod
    def get_module(cls, _host: "testinfra.host.Host") -> type["Module"]:
        klass = cls.get_module_class(_host)
        return type(
            klass.__name__,
            (klass,),
            {"_host": _host},
        )

    @classmethod
    def get_module_class(cls, host):
        return cls

    @classmethod
    def run(cls, *args, **kwargs):
        return cls._host.run(*args, **kwargs)

    @classmethod
    def run_test(cls, *args, **kwargs):
        return cls._host.run_test(*args, **kwargs)

    @classmethod
    def run_expect(cls, *args, **kwargs):
        return cls._host.run_expect(*args, **kwargs)

    @classmethod
    def check_output(cls, *args, **kwargs):
        return cls._host.check_output(*args, **kwargs)

    @classmethod
    def find_command(cls, *args, **kwargs):
        return cls._host.find_command(*args, **kwargs)


class InstanceModule(Module):
    @classmethod
    def get_module(cls, _host):
        klass = super().get_module(_host)
        return klass()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

from testinfra.modules.base import InstanceModule


class Sudo(InstanceModule):
    """Sudo module allow to run certain portion of code under another user.

    It is used as a context manager and can be nested.

    >>> Command.check_output("whoami")
    'phil'
    >>> with host.sudo():
    ...     host.check_output("whoami")
    ...     with host.sudo("www-data"):
    ...         host.check_output("whoami")
    ...
    'root'
    'www-data'

    """

    @contextlib.contextmanager
    def __call__(self, user=None):
        old_get_command = self._host.backend.get_command
        quote = self._host.backend.quote
        get_sudo_command = self._host.backend.get_sudo_command

        def get_command(command, *args):
            return old_get_command(get_sudo_command(quote(command, *args), user))

        self._host.backend.get_command = get_command
        try:
            yield
        finally:
            self._host.backend.get_command = old_get_command

    def __repr__(self):
        return "<sudo>"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

from testinfra.modules.base import Module


class Package(Module):
    """Test packages status and version"""

    def __init__(self, name):
        self.name = name
        super().__init__()

    @property
    def is_installed(self):
        """Test if the package is installed

        >>> host.package("nginx").is_installed
        True

        Supported package systems:

        - apk (Alpine)
        - apt (Debian, Ubuntu, ...)
        - brew (macOS)
        - pacman (Arch, Manjaro )
        - pkg (FreeBSD)
        - pkg_info (NetBSD)
        - pkg_info (OpenBSD)
        - rpm (RHEL, RockyLinux, Fedora, ...)
        """
        raise NotImplementedError

    @property
    def release(self):
        """Return the release specific info from the package version

        >>> host.package("nginx").release
        '1.el6'
        """
        raise NotImplementedError

    @property
    def version(self):
        """Return package version as returned by the package system

        >>> host.package("nginx").version
        '1.2.1-2.2+wheezy3'
        """
        raise NotImplementedError

    def __repr__(self):
        return f"<package {self.name}>"

    @classmethod
    def get_module_class(cls, host):
        if host.system_info.type == "windows":
            return ChocolateyPackage
        if host.system_info.type == "freebsd":
            return FreeBSDPackage
        if host.system_info.type in ("openbsd", "netbsd"):
            return OpenBSDPackage
        if host.system_info.distribution in ("debian", "ubuntu"):
            return DebianPackage
        if host.system_info.distribution and (
            host.system_info.distribution.lower()
            in (
                "almalinux",
                "centos",
                "cloudlinux",
                "fedora",
                "ol",
                "opensuse-leap",
                "opensuse-tumbleweed",
                "rhel",
                "rocky",
            )
        ):
            return RpmPackage
        if host.system_info.distribution in ("arch", "manjarolinux"):
            return ArchPackage
        if host.exists("apk"):
            return AlpinePackage
        # Fallback conditions
        if host.exists("dpkg-query"):
            return DebianPackage
        if host.exists("rpm"):
            return RpmPackage
        if host.exists("brew"):
            return HomebrewPackage
        raise NotImplementedError


class DebianPackage(Package):
    @property
    def is_installed(self):
        result = self.run_test("dpkg-query -f '${Status}' -W %s", self.name)
        if result.rc == 1:
            return False
        out = result.stdout.strip().split()
        installed_status = ["ok", "installed"]
        return out[0] in ["install", "hold"] and out[1:3] == installed_status

    @property
    def release(self):
        raise NotImplementedError

    @property
    def version(self):
        out = self.check_output("dpkg-query -f '${Status} ${Version}' -W %s", self.name)
        splitted = out.split()
        assert splitted[0].lower() in (
            "install",
            "hold",
        ), f"The package {self.name} is not installed, dpkg-query output: {out}"
        return splitted[3]


class FreeBSDPackage(Package):
    @property
    def is_installed(self):
        EX_UNAVAILABLE = 69
        return (
            self.run_expect([0, EX_UNAVAILABLE], "pkg query %%n %s", self.name).rc == 0
        )

    @property
    def release(self):
        raise NotImplementedError

    @property
    def version(self):
        return self.check_output("pkg query %%v %s", self.name)


class OpenBSDPackage(Package):
    @property
    def is_installed(self):
        return self.run_test("pkg_info -e %s", f"{self.name}-*").rc == 0

    @property
    def release(self):
        raise NotImplementedError

    @property
    def version(self):
        out = self.check_output("pkg_info -e %s", f"{self.name}-*")
        # OpenBSD: inst:zsh-5.0.5p0
        # NetBSD: zsh-5.0.7nb1
        return out.split(self.name + "-", 1)[1]


class RpmPackage(Package):
    @property
    def is_installed(self):
        result = self.run_test("rpm -q --quiet %s 2>&1", self.name)
        if result.succeeded:
            return True
        elif result.failed and result.stdout == "":
            return False
        else:
            raise RuntimeError(
                f"Could not check if RPM package '{self.name}' is installed. {result.stdout}"
            )

    @property
    def version(self):
        return self.check_output('rpm -q --queryformat="%%{VERSION}" %s', self.name)

    @property
    def release(self):
        return self.check_output('rpm -q --queryformat="%%{RELEASE}" %s', self.name)


class AlpinePackage(Package):
    @property
    def is_installed(self):
        return self.run_test("apk -e info %s", self.name).rc == 0

    @property
    def version(self):
        out = self.check_output("apk -e -v info %s", self.name).split("-")
        return out[-2]

    @property
    def release(self):
        out = self.check_output("apk -e -v info %s", self.name).split("-")
        return out[-1]


class ArchPackage(Package):
    @property
    def is_installed(self):
        return self.run_test("pacman -Q %s", self.name).rc == 0

    @property
    def version(self):
        out = self.check_output("pacman -Q %s", self.name).split(" ")
        return out[1]

    @property
    def release(self):
        raise NotImplementedError


class ChocolateyPackage(Package):
    @property
    def is_installed(self):
        return self.run_test("choco info -lo %s", self.name).rc == 0

    @property
    def version(self):
        _, version = self.check_output("choco info -lo %s -r", self.name).split("|", 1)
        return version

    @property
    def release(self):
        raise NotImplementedError


class HomebrewPackage(Package):
    @property
    def is_installed(self):
        info = self.check_output("brew info --formula --json %s", self.name)
        return len(json.loads(info)[0]["installed"]) > 0

    @property
    def version(self):
        info = self.check_output("brew info --formula --json %s", self.name)
        version = json.loads(info)[0]["installed"][0]["version"]
        return version

    @property
    def release(self):
        raise NotImplementedError



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from testinfra.modules.base import InstanceModule


def parse_puppet_resource(data: str) -> dict[str, dict[str, str]]:
    """Parse data returned by 'puppet resource'

    $ puppet resource user
    user { 'root':
        ensure  => 'present',
        comment => 'root',
        gid     => '0',
        home    => '/root',
        shell   => '/usr/bin/zsh',
        uid     => '0',
    }
    user { 'sshd':
      ensure => 'present',
      gid    => '65534',
      home   => '/var/run/sshd',
      shell  => '/usr/sbin/nologin',
      uid    => '106',
    }
    [...]
    """

    state: dict[str, dict[str, str]] = {}
    current = None
    for line in data.splitlines():
        if not current:
            current = line.split("'")[1]
            state[current] = {}
        elif current and line == "}":
            current = None
        elif current:
            key, value = line.split(" => ")
            key = key.strip()
            value = value.split("'")[1]
            state[current][key] = value
    return state


class PuppetResource(InstanceModule):
    """Get puppet resources

    Run ``puppet resource --types`` to get a list of available types.

    >>> host.puppet_resource("user", "www-data")
    {
        'www-data': {
            'ensure': 'present',
            'comment': 'www-data',
            'gid': '33',
            'home': '/var/www',
            'shell': '/usr/sbin/nologin',
            'uid': '33',
        },
    }
    """

    def __call__(self, resource_type, name=None):
        cmd = "puppet resource %s"
        args = [resource_type]
        if name is not None:
            cmd += " %s"
            args.append(name)
        # TODO(phil): Since puppet 4.0.0 puppet resource has a --to_yaml option
        return parse_puppet_resource(self.check_output(cmd, *args))

    def __repr__(self):
        return "<PuppetResource>"


class Facter(InstanceModule):
    """Get facts with `facter <https://puppetlabs.com/facter>`_

    >>> host.facter()
    {
        "operatingsystem": "Debian",
        "kernel": "linux",
        [...]
    }
    >>> host.facter("kernelversion", "is_virtual")
    {
      "kernelversion": "3.16.0",
      "is_virtual": "false"
    }
    """

    def __call__(self, *facts):
        cmd = "facter --json --puppet " + " ".join(facts)
        return json.loads(self.check_output(cmd))

    def __repr__(self):
        return "<facter>"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import pprint
from typing import Any, Optional

from testinfra.backend import base
from testinfra.utils.ansible_runner import AnsibleRunner

logger = logging.getLogger("testinfra")


class AnsibleBackend(base.BaseBackend):
    NAME = "ansible"
    HAS_RUN_ANSIBLE = True

    def __init__(
        self,
        host: str,
        ansible_inventory: Optional[str] = None,
        ssh_config: Optional[str] = None,
        ssh_identity_file: Optional[str] = None,
        force_ansible: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        self.host = host
        self.ansible_inventory = ansible_inventory
        self.ssh_config = ssh_config
        self.ssh_identity_file = ssh_identity_file
        self.force_ansible = force_ansible
        super().__init__(host, *args, **kwargs)

    @property
    def ansible_runner(self) -> AnsibleRunner:
        return AnsibleRunner.get_runner(self.ansible_inventory)

    def run(self, command: str, *args: str, **kwargs: Any) -> base.CommandResult:
        command = self.get_command(command, *args)
        if not self.force_ansible:
            host = self.ansible_runner.get_host(
                self.host,
                ssh_config=self.ssh_config,
                ssh_identity_file=self.ssh_identity_file,
            )
            if host is not None:
                return host.run(command)
        out = self.run_ansible("shell", module_args=command, check=False)
        if "module_stdout" in out:
            data = json.loads(out["module_stdout"])
            stdout = data["stdout"]
            stderr = data["stderr"]
        else:
            # bw compat
            stdout = out["stdout"]
            stderr = out["stderr"]
        return self.result(out["rc"], self.encode(command), stdout, stderr)

    def run_ansible(
        self, module_name: str, module_args: Optional[str] = None, **kwargs: Any
    ) -> Any:
        def get_encoding() -> str:
            return self.encoding

        result = self.ansible_runner.run_module(
            self.host, module_name, module_args, get_encoding=get_encoding, **kwargs
        )
        logger.info(
            "RUN Ansible(%s, %s, %s): %s",
            repr(module_name),
            repr(module_args),
            repr(kwargs),
            pprint.pformat(result),
        )
        return result

    def get_variables(self) -> dict[str, Any]:
        return self.ansible_runner.get_variables(self.host)

    @classmethod
    def get_hosts(cls, host: str, **kwargs: Any) -> list[str]:
        inventory = kwargs.get("ansible_inventory")
        return AnsibleRunner.get_runner(inventory).get_hosts(host or "all")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testinfra.modules.base import Module


class _AddrPort:
    def __init__(self, addr, port):
        self._addr = addr
        self._port = str(port)

    @property
    def is_reachable(self):
        """Return if port is reachable"""
        if not self._addr._host.exists("nc"):
            if self._addr.namespace:
                # in this case cannot use namespace
                raise NotImplementedError(
                    "nc command not available, namespace cannot be used"
                )
            # Fallback to bash if netcat is not available
            return (
                self._addr.run_expect(
                    [0, 1, 124],
                    "timeout 1 bash -c 'cat < /dev/null > /dev/tcp/%s/%s'",
                    self._addr.name,
                    self._port,
                ).rc
                == 0
            )

        return (
            self._addr.run(
                f"{self._addr._prefix}nc -w 1 -z {self._addr.name} {self._port}"
            ).rc
            == 0
        )


class Addr(Module):
    """Test remote address

    Example:

    >>> google = host.addr("google.com")
    >>> google.is_resolvable
    True
    >>> '173.194.32.225' in google.ipv4_addresses
    True
    >>> google.is_reachable
    True
    >>> google.port(443).is_reachable
    True
    >>> google.port(666).is_reachable
    False

    Can also be use within a network namespace_.

    >>> localhost = host.addr("localhost", "ns1")
    >>> localhost.is_resolvable
    True

    Network namespaces can only be used if ip_ command is available
    because in this case, the module use ip-netns_ as command prefix.
    In the other case, it will raise NotImplementedError.

    .. _namespace: https://man7.org/linux/man-pages/man7/namespaces.7.html
    .. _ip: https://man7.org/linux/man-pages/man8/ip.8.html
    .. _ip-netns: https://man7.org/linux/man-pages/man8/ip-netns.8.html
    """

    def __init__(self, name, namespace=None):
        self._name = name
        self._namespace = namespace
        if self.namespace and not self._host.exists("ip"):
            raise NotImplementedError(
                "ip command not available, namespace cannot be used"
            )
        super().__init__()

    @property
    def name(self):
        """Return host name"""
        return self._name

    @property
    def namespace(self):
        """Return network namespace"""
        return self._namespace

    @property
    def _prefix(self):
        """Return the prefix to use for commands"""
        prefix = ""
        if self.namespace:
            prefix = f"ip netns exec {self.namespace} "
        return prefix

    @property
    def namespace_exists(self):
        """Test if the network namespace exists"""
        # could use ip netns list instead
        return (
            self.namespace
            and self.run_test("test -e /var/run/netns/%s", self.namespace).rc == 0
        )

    @property
    def is_resolvable(self):
        """Return if address is resolvable"""
        return len(self.ip_addresses) > 0

    @property
    def is_reachable(self):
        """Return if address is reachable"""
        return (
            self.run_expect([0, 1, 2], f"{self._prefix}ping -W 1 -c 1 {self.name}").rc
            == 0
        )

    @property
    def ip_addresses(self):
        """Return IP addresses of host"""
        return self._resolve("ahosts")

    @property
    def ipv4_addresses(self):
        """Return IPv4 addresses of host"""
        return self._resolve("ahostsv4")

    @property
    def ipv6_addresses(self):
        """Return IPv6 addresses of host"""
        return self._resolve("ahostsv6")

    def port(self, port):
        """Return address-port pair"""
        return _AddrPort(self, port)

    def __repr__(self):
        return f"<addr {self.name}>"

    def _resolve(self, method):
        result = self.run_expect(
            [0, 1, 2], f"{self._prefix}getent {method} {self.name}"
        )
        lines = result.stdout.splitlines()
        return list({line.split()[0] for line in lines})



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testinfra.modules.base import Module


class Group(Module):
    """Test unix group"""

    def __init__(self, name):
        self.name = name
        super().__init__()

    @property
    def exists(self):
        """Test if group exists

        >>> host.group("wheel").exists
        True
        >>> host.group("nosuchgroup").exists
        False
        """
        return self.run_expect([0, 2], "getent group %s", self.name).rc == 0

    @property
    def get_all_groups(self):
        """Returns a list of local and remote group names

        >>> host.group("anyname").get_all_groups
        ["root", "wheel", "man", "tty", <...>]
        """
        all_groups = [
            line.split(":")[0]
            for line in self.check_output("getent group").splitlines()
        ]
        return all_groups

    @property
    def get_local_groups(self):
        """Returns a list of local group names

        >>> host.group("anyname").get_local_groups
        ["root", "wheel", "man", "tty", <...>]
        """
        local_groups = [
            line.split(":")[0]
            for line in self.check_output("cat /etc/group").splitlines()
        ]
        return local_groups

    @property
    def gid(self):
        return int(self.check_output("getent group %s | cut -d':' -f3", self.name))

    @property
    def members(self):
        """Return all users that are members of this group."""
        users = self.check_output("getent group %s | cut -d':' -f4", self.name)
        if users:
            return users.split(",")
        return []

    def __repr__(self):
        return f"<group {self.name}>"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from testinfra.modules.base import Module


class Service(Module):
    """Test services

    Implementations:

    - Linux: detect Systemd, Upstart or OpenRC, fallback to SysV
    - FreeBSD: service(1)
    - OpenBSD: ``/etc/rc.d/$name check`` for ``is_running``
      ``rcctl ls on`` for ``is_enabled`` (only OpenBSD >= 5.8)
    - NetBSD: ``/etc/rc.d/$name onestatus`` for ``is_running``
      (``is_enabled`` is not yet implemented)

    """

    def __init__(self, name):
        self.name = name
        super().__init__()

    @property
    def exists(self):
        """Test if service is exists"""
        raise NotImplementedError

    @property
    def is_running(self):
        """Test if service is running"""
        raise NotImplementedError

    @property
    def is_enabled(self):
        """Test if service is enabled"""
        raise NotImplementedError

    @property
    def is_valid(self):
        """Test if service is valid

        This method is only available in the systemd implementation,
        it will raise ``NotImplementedError`` in others implementation
        """
        raise NotImplementedError

    @property
    def is_masked(self):
        """Test if service is masked

        This method is only available in the systemd implementation,
        it will raise ``NotImplementedError`` in others implementations
        """
        raise NotImplementedError

    @functools.cached_property
    def systemd_properties(self):
        """Properties of the service (unit).

        Return service properties as a `dict`,
        empty properties are not returned.

        >>> ntp = host.service("ntp")
        >>> ntp.systemd_properties["FragmentPath"]
        '/lib/systemd/system/ntp.service'

        This method is only available in the systemd implementation,
        it will raise ``NotImplementedError`` in others implementations

        Note: based on `systemctl show`_

        .. _systemctl show: https://man7.org/linux/man-pages/man1/systemctl.1.html
        """
        raise NotImplementedError

    @classmethod
    def get_module_class(cls, host):
        if host.system_info.type == "linux":
            if host.file("/run/systemd/system/").is_directory or (
                host.exists("systemctl")
                and "systemd" in host.file("/sbin/init").linked_to
            ):
                return SystemdService
            if (
                host.exists("initctl")
                and host.exists("status")
                and host.file("/etc/init").is_directory
            ):
                return UpstartService
            if host.exists("rc-service"):
                return OpenRCService
            return SysvService
        if host.system_info.type == "freebsd":
            return FreeBSDService
        if host.system_info.type == "openbsd":
            return OpenBSDService
        if host.system_info.type == "netbsd":
            return NetBSDService
        if host.system_info.type == "windows":
            return WindowsService
        raise NotImplementedError

    def __repr__(self):
        return f"<service {self.name}>"


class SysvService(Service):
    @functools.cached_property
    def _service_command(self):
        return self.find_command("service")

    @property
    def is_running(self):
        # based on /lib/lsb/init-functions
        # 0: program running
        # 1: program is dead and pid file exists
        # 3: not running and pid file does not exists
        # 4: Unable to determine status
        # 8: starting (alpine specific ?)
        return (
            self.run_expect(
                [0, 1, 3, 8], "%s %s status", self._service_command, self.name
            ).rc
            == 0
        )

    @property
    def is_enabled(self):
        return bool(
            self.check_output(
                "find -L /etc/rc?.d/ -name %s",
                "S??" + self.name,
            )
        )


class SystemdService(SysvService):
    suffix_list = [
        "service",
        "socket",
        "device",
        "mount",
        "automount",
        "swap",
        "target",
        "path",
        "timer",
        "slice",
        "scope",
    ]
    """
    List of valid suffixes for systemd unit files

    See systemd.unit(5) for more details
    """

    def _has_systemd_suffix(self):
        """
        Check if service name has a known systemd unit suffix
        """
        unit_suffix = self.name.split(".")[-1]
        return unit_suffix in self.suffix_list

    @property
    def exists(self):
        cmd = self.run_test('systemctl list-unit-files | grep -q "^%s"', self.name)
        return cmd.rc == 0

    @property
    def is_running(self):
        # based on https://man7.org/linux/man-pages/man1/systemctl.1.html
        # 0: program running
        # 1: program is dead and pid file exists
        # 3: not running and pid file does not exists
        # 4: Unable to determine status (no such unit)
        out = self.run_expect([0, 1, 3, 4], "systemctl is-active %s", self.name)
        if out.rc == 1:
            # Failed to connect to bus: No such file or directory
            return super().is_running
        return out.rc == 0

    @property
    def is_enabled(self):
        cmd = self.run_test("systemctl is-enabled %s", self.name)
        if cmd.rc == 0:
            return True
        if cmd.stdout.strip() == "disabled":
            return False
        # Fallback on SysV - only for non-systemd units
        # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=760616
        if not self._has_systemd_suffix():
            return super().is_enabled
        raise RuntimeError(
            f"Unable to determine state of {self.name}. Does this service exist?"
        )

    @property
    def is_valid(self):
        # systemd-analyze requires a full unit name.
        name = self.name if self._has_systemd_suffix() else f"{self.name}.service"
        cmd = self.run("systemd-analyze verify %s", name)
        # A bad unit file still returns a rc of 0, so check the
        # stdout for anything.  Nothing means no warns/errors.
        # Docs at https://www.freedesktop.org/software/systemd/man/systemd
        # -analyze.html#Examples%20for%20verify
        assert (cmd.stdout, cmd.stderr) == ("", "")
        return True

    @property
    def is_masked(self):
        cmd = self.run_test("systemctl is-enabled %s", self.name)
        return cmd.stdout.strip() == "masked"

    @functools.cached_property
    def systemd_properties(self):
        out = self.check_output("systemctl show %s", self.name)
        out_d = {}
        if out:
            # maxsplit is required because values can contain `=`
            out_d = dict(
                map(lambda pair: pair.split("=", maxsplit=1), out.splitlines())
            )
        return out_d


class UpstartService(SysvService):
    @property
    def exists(self):
        return self._host.file(f"/etc/init/{self.name}.conf").exists

    @property
    def is_enabled(self):
        if (
            self.run(
                "grep -q '^start on' /etc/init/%s.conf",
                self.name,
            ).rc
            == 0
            and self.run(
                "grep -q '^manual' /etc/init/%s.override",
                self.name,
            ).rc
            != 0
        ):
            return True
        # Fallback on SysV
        return super().is_enabled

    @property
    def is_running(self):
        cmd = self.run_test("status %s", self.name)
        if cmd.rc == 0 and len(cmd.stdout.split()) > 1:
            return "running" in cmd.stdout.split()[1]
        return super().is_running


class OpenRCService(SysvService):
    @functools.cached_property
    def _service_command(self):
        return self.find_command("rc-service")

    @property
    def is_enabled(self):
        return bool(
            self.check_output(
                "find /etc/runlevels/ -name %s",
                self.name,
            )
        )


class FreeBSDService(Service):
    @property
    def exists(self):
        return self._host.file(f"/etc/rc.d/{self.name}").exists

    @property
    def is_running(self):
        return self.run_test("service %s onestatus", self.name).rc == 0

    @property
    def is_enabled(self):
        # Return list of enabled services like
        # /etc/rc.d/sshd
        # /etc/rc.d/sendmail
        for path in self.check_output("service -e").splitlines():
            if path and path.rsplit("/", 1)[1] == self.name:
                return True
        return False


class OpenBSDService(Service):
    @property
    def exists(self):
        return self._host.file(f"/etc/rc.d/{self.name}").exists

    @property
    def is_running(self):
        return self.run_test("/etc/rc.d/%s check", self.name).rc == 0

    @property
    def is_enabled(self):
        if self.name in self.check_output("rcctl ls on").splitlines():
            return True
        if self.name in self.check_output("rcctl ls off").splitlines():
            return False
        raise RuntimeError(
            f"Unable to determine state of {self.name}. Does this service exist?"
        )


class NetBSDService(Service):
    @property
    def exists(self):
        return self._host.file(f"/etc/rc.d/{self.name}").exists

    @property
    def is_running(self):
        return self.run_test("/etc/rc.d/%s onestatus", self.name).rc == 0

    @property
    def is_enabled(self):
        raise NotImplementedError


class WindowsService(Service):
    @property
    def exists(self):
        out = self.check_output(
            f"Get-Service -Name {self.name} -ErrorAction SilentlyContinue"
        )
        return self.name in out

    @property
    def is_running(self):
        return (
            self.check_output(
                "Get-Service '%s' | Select -ExpandProperty Status",
                self.name,
            )
            == "Running"
        )

    @property
    def is_enabled(self):
        return (
            self.check_output(
                "Get-Service '%s' | Select -ExpandProperty StartType",
                self.name,
            )
            == "Automatic"
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NoReturn

from testinfra.modules.base import Module

STATUS = [
    "STOPPED",
    "STARTING",
    "RUNNING",
    "BACKOFF",
    "STOPPING",
    "EXITED",
    "FATAL",
    "UNKNOWN",
]


def supervisor_not_running() -> NoReturn:
    raise RuntimeError("Cannot get supervisor status. Is supervisor running ?")


class Supervisor(Module):
    """Test supervisor managed services

    >>> gunicorn = host.supervisor("gunicorn")
    >>> gunicorn.status
    'RUNNING'
    >>> gunicorn.is_running
    True
    >>> gunicorn.pid
    4242

    The path where supervisorctl and its configuration file reside can be specified.

    >>> gunicorn = host.supervisor("gunicorn", "/usr/bin/supervisorctl", "/etc/supervisor/supervisord.conf")
    >>> gunicorn.status
    'RUNNING'
    """

    def __init__(
        self,
        name,
        supervisorctl_path="supervisorctl",
        supervisorctl_conf=None,
        _attrs_cache=None,
    ):
        self.name = name
        self.supervisorctl_path = supervisorctl_path
        self.supervisorctl_conf = supervisorctl_conf
        self._attrs_cache = _attrs_cache
        super().__init__()

    @staticmethod
    def _parse_status(line):
        splitted = line.split()
        name = splitted[0]
        status = splitted[1]
        # some old supervisorctl versions exit status is 0 even if it cannot
        # connect to supervisord socket and output the error to stdout.  So we
        # check that parsed status is a known status.
        if status not in STATUS:
            supervisor_not_running()
        pid = int(splitted[3].removesuffix(",")) if status == "RUNNING" else None
        return {"name": name, "status": status, "pid": pid}

    @property
    def _attrs(self):
        if self._attrs_cache is None:
            if self.supervisorctl_conf:
                out = self.run_expect(
                    [0, 3, 4],
                    "%s -c %s status %s",
                    self.supervisorctl_path,
                    self.supervisorctl_conf,
                    self.name,
                )
            else:
                out = self.run_expect(
                    [0, 3, 4], "%s status %s", self.supervisorctl_path, self.name
                )
            if out.rc == 4:
                supervisor_not_running()
            line = out.stdout.rstrip("\r\n")
            attrs = self._parse_status(line)
            assert attrs["name"] == self.name
            self._attrs_cache = attrs
        return self._attrs_cache

    @property
    def is_running(self):
        """Return True if managed service is in status RUNNING"""
        return self.status == "RUNNING"

    @property
    def status(self):
        """Return the status of the managed service

        Status can be STOPPED, STARTING, RUNNING, BACKOFF, STOPPING,
        EXITED, FATAL, UNKNOWN.

        See http://supervisord.org/subprocess.html#process-states
        """
        return self._attrs["status"]

    @property
    def pid(self):
        """Return the pid (as int) of the managed service"""
        return self._attrs["pid"]

    @classmethod
    def get_services(
        cls,
        supervisorctl_path="supervisorctl",
        supervisorctl_conf=None,
    ):
        """Get a list of services running under supervisor

        >>> host.supervisor.get_services()
        [<Supervisor(name="gunicorn", status="RUNNING", pid=4232)>
         <Supervisor(name="celery", status="FATAL", pid=None)>]

        The path where supervisorctl and its configuration file reside can be specified.

        >>> host.supervisor.get_services("/usr/bin/supervisorctl", "/etc/supervisor/supervisord.conf")
        [<Supervisor(name="gunicorn", status="RUNNING", pid=4232)>
         <Supervisor(name="celery", status="FATAL", pid=None)>]
        """
        services = []
        if supervisorctl_conf:
            out = cls.check_output(
                "%s -c %s status", supervisorctl_path, supervisorctl_conf
            )
        else:
            out = cls.check_output("%s status", supervisorctl_path)
        for line in out.splitlines():
            attrs = cls._parse_status(line)
            service = cls(
                attrs["name"],
                supervisorctl_path=supervisorctl_path,
                supervisorctl_conf=supervisorctl_conf,
                _attrs_cache=attrs,
            )
            services.append(service)
        return services

    def __repr__(self):
        return f"<Supervisor(name={self.name}, status={self.status}, pid={self.pid})>"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import dataclasses
import locale
import logging
import shlex
import subprocess
import urllib.parse
from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    import testinfra.host

logger = logging.getLogger("testinfra")


@dataclasses.dataclass
class HostSpec:
    name: str
    port: Optional[str]
    user: Optional[str]
    password: Optional[str]


@dataclasses.dataclass
class CommandResult:
    """Object that encapsulates all returned details of the command execution.

    Example:

    >>> cmd = host.run("ls -l /etc/passwd")
    >>> cmd.rc
    0
    >>> cmd.stdout
    '-rw-r--r-- 1 root root 1790 Feb 11 00:28 /etc/passwd\\n'
    >>> cmd.stderr
    ''
    >>> cmd.succeeded
    True
    >>> cmd.failed
    False
    """

    backend: "BaseBackend"
    exit_status: int
    command: bytes
    _stdout: Union[str, bytes]
    _stderr: Union[str, bytes]

    @property
    def succeeded(self) -> bool:
        """Returns whether the command was successful

        >>> host.run("true").succeeded
        True
        """
        return self.exit_status == 0

    @property
    def failed(self) -> bool:
        """Returns whether the command failed

        >>> host.run("false").failed
        True
        """
        return self.exit_status != 0

    @property
    def rc(self) -> int:
        """Gets the returncode of a command

        >>> host.run("true").rc
        0
        """
        return self.exit_status

    @property
    def stdout(self) -> str:
        """Gets standard output (stdout) stream of an executed command

        >>> host.run("mkdir -v new_directory").stdout
        mkdir: created directory 'new_directory'
        """
        if isinstance(self._stdout, bytes):
            return self.backend.decode(self._stdout)
        return self._stdout

    @property
    def stderr(self) -> str:
        """Gets standard error (stderr) stream of an executed command

        >>> host.run("mkdir new_directory").stderr
        mkdir: cannot create directory 'new_directory': File exists
        """
        if isinstance(self._stderr, bytes):
            return self.backend.decode(self._stderr)
        return self._stderr

    @property
    def stdout_bytes(self) -> bytes:
        """Gets standard output (stdout) stream of an executed command as bytes

        >>> host.run("mkdir -v new_directory").stdout_bytes
        b"mkdir: created directory 'new_directory'"
        """
        if isinstance(self._stdout, str):
            return self.backend.encode(self._stdout)
        return self._stdout

    @property
    def stderr_bytes(self) -> bytes:
        """Gets standard error (stderr) stream of an executed command as bytes

        >>> host.run("mkdir new_directory").stderr_bytes
        b"mkdir: cannot create directory 'new_directory': File exists"
        """
        if isinstance(self._stderr, str):
            return self.backend.encode(self._stderr)
        return self._stderr


class BaseBackend(metaclass=abc.ABCMeta):
    """Represent the connection to the remote or local system"""

    HAS_RUN_SALT = False
    HAS_RUN_ANSIBLE = False
    NAME: str

    def __init__(
        self,
        hostname: str,
        sudo: bool = False,
        sudo_user: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ):
        self._encoding: Optional[str] = None
        self._host: Optional[testinfra.host.Host] = None
        self.hostname = hostname
        self.sudo = sudo
        self.sudo_user = sudo_user
        super().__init__()

    def set_host(self, host: "testinfra.host.Host") -> None:
        self._host = host

    @classmethod
    def get_connection_type(cls) -> str:
        """Return the connection backend used as string.

        Can be local, paramiko, ssh, docker, salt or ansible
        """
        return cls.NAME

    def get_hostname(self) -> str:
        """Return the hostname (for testinfra) of the remote or local system


        Can be useful for multi-hosts tests:

        Example:
        ::

            import requests


            def test(TestinfraBackend):
                host = TestinfraBackend.get_hostname()
                response = requests.get("http://" + host)
                assert response.status_code == 200


        ::

            $ testinfra --hosts=server1,server2 test.py

            test.py::test[paramiko://server1] PASSED
            test.py::test[paramiko://server2] PASSED
        """
        return self.hostname

    def get_pytest_id(self) -> str:
        return self.get_connection_type() + "://" + self.get_hostname()

    @classmethod
    def get_hosts(cls, host: str, **kwargs: Any) -> list[str]:
        if host is None:
            raise RuntimeError(
                f"One or more hosts is required with the {cls.get_connection_type()} backend"
            )
        return [host]

    @staticmethod
    def quote(command: str, *args: str) -> str:
        if args:
            return command % tuple(shlex.quote(a) for a in args)
        return command

    def get_sudo_command(self, command: str, sudo_user: Optional[str]) -> str:
        if sudo_user is None:
            return self.quote("sudo /bin/sh -c %s", command)
        return self.quote("sudo -u %s /bin/sh -c %s", sudo_user, command)

    def get_command(self, command: str, *args: str) -> str:
        command = self.quote(command, *args)
        if self.sudo:
            command = self.get_sudo_command(command, self.sudo_user)
        return command

    def run(self, command: str, *args: str, **kwargs: Any) -> CommandResult:
        raise NotImplementedError

    def run_local(self, command: str, *args: str) -> CommandResult:
        command = self.quote(command, *args)
        cmd = self.encode(command)
        p = subprocess.Popen(
            cmd,
            shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate()
        result = self.result(p.returncode, cmd, stdout, stderr)
        return result

    @staticmethod
    def parse_hostspec(hostspec: str) -> HostSpec:
        name = hostspec
        port = None
        user = None
        password = None
        if "@" in name:
            user, name = name.split("@", 1)
            if ":" in user:
                user, password = user.split(":", 1)
        # A literal IPv6 address might be like
        #  [fe80:0::a:b:c]:80
        # thus, below in words; if this starts with a '[' assume it
        # encloses an ipv6 address with a closing ']', with a possible
        # trailing port after a colon
        if name.startswith("["):
            name, port = name.split("]")
            name = name[1:]
            port = port[1:] if port.startswith(":") else None
        else:
            if ":" in name:
                name, port = name.split(":", 1)
        name = urllib.parse.unquote(name)
        if user is not None:
            user = urllib.parse.unquote(user)
        if password is not None:
            password = urllib.parse.unquote(password)
        return HostSpec(name, port, user, password)

    @staticmethod
    def parse_containerspec(containerspec: str) -> tuple[str, Optional[str]]:
        name = containerspec
        user = None
        if "@" in name:
            user, name = name.split("@", 1)
        return name, user

    def get_encoding(self) -> str:
        encoding = None
        for python in ("python3", "python"):
            cmd = self.run(
                "%s -c 'import locale;print(locale.getpreferredencoding())'",
                python,
                encoding=None,
            )
            if cmd.rc == 0:
                encoding = cmd.stdout_bytes.splitlines()[0].decode("ascii")
                break
        # Python is not installed, we hope the encoding to be the same as
        # local machine...
        if not encoding:
            encoding = locale.getpreferredencoding()
        if encoding == "ANSI_X3.4-1968":
            # Workaround default encoding ascii without LANG set
            encoding = "UTF-8"
        return encoding

    @property
    def encoding(self) -> str:
        if self._encoding is None:
            self._encoding = self.get_encoding()
        return self._encoding

    def decode(self, data: bytes) -> str:
        try:
            return data.decode("ascii")
        except UnicodeDecodeError:
            return data.decode(self.encoding)

    def encode(self, data: str) -> bytes:
        try:
            return data.encode("ascii")
        except UnicodeEncodeError:
            return data.encode(self.encoding)

    def result(
        self, rc: int, cmd: bytes, stdout: Union[str, bytes], stderr: Union[str, bytes]
    ) -> CommandResult:
        result = CommandResult(
            backend=self,
            exit_status=rc,
            command=cmd,
            _stdout=stdout,
            _stderr=stderr,
        )
        logger.debug("RUN %s", result)
        return result



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import pprint

from testinfra.modules.base import InstanceModule


class AnsibleException(Exception):
    """Exception raised when an error occur in an ansible call

    result from ansible can be accessed through the ``result`` attribute

    >>> try:
    ...     host.ansible("command", "echo foo")
    ... except host.ansible.AnsibleException as exc:
    ...     assert exc.result['failed'] is True
    ...     assert exc.result['msg'] == 'Skipped. You might want to try check=False'  # noqa
    """

    def __init__(self, result):
        self.result = result
        super().__init__(f"Unexpected error: {pprint.pformat(result)}")


def need_ansible(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self._host.backend.HAS_RUN_ANSIBLE:
            raise RuntimeError(
                "Ansible module is only available with ansible connection backend"
            )
        return func(self, *args, **kwargs)

    return wrapper


class Ansible(InstanceModule):
    """Run Ansible module functions

    This module is only available with the :ref:`ansible connection
    backend` connection backend.

    `Check mode
    <https://docs.ansible.com/ansible/playbooks_checkmode.html>`_ is
    enabled by default, you can disable it with `check=False`.

    `Become
    <https://docs.ansible.com/ansible/user_guide/become.html#id1>`_ is
    `False` by default. You can enable it with `become=True`.

    Ansible arguments that are not related to the Ansible inventory or
    connection (both managed by testinfra) are also accepted through keyword
    arguments:

        - ``become_method`` *str* sudo, su, doas, etc.
        - ``become_user`` *str* become this user.
        - ``diff`` *bool*: when changing (small) files and templates, show the
          differences in those files.
        - ``extra_vars`` *dict* serialized to a JSON string, passed to
          Ansible.
        - ``one_line`` *bool*: condense output.
        - ``user`` *str* connect as this user.
        - ``verbose`` *int* level of verbosity

    >>> host.ansible("apt", "name=nginx state=present")["changed"]
    False
    >>> host.ansible("apt", "name=nginx state=present", become=True)["changed"]
    False
    >>> host.ansible("command", "echo foo", check=False)["stdout"]
    'foo'
    >>> host.ansible("setup")["ansible_facts"]["ansible_lsb"]["codename"]
    'jessie'
    >>> host.ansible("file", "path=/etc/passwd")["mode"]
    '0640'
    >>> host.ansible(
    ... "command",
    ... "id --user --name",
    ... check=False,
    ... become=True,
    ... become_user="http",
    ... )["stdout"]
    'http'
    >>> host.ansible(
    ... "apt",
    ... "name={{ packages }}",
    ... check=False,
    ... extra_vars={"packages": ["neovim", "vim"]},
    ... )
    # Installs neovim and vim.

    """

    AnsibleException = AnsibleException

    @need_ansible
    def __call__(
        self, module_name, module_args=None, check=True, become=False, **kwargs
    ):
        result = self._host.backend.run_ansible(
            module_name, module_args, check=check, become=become, **kwargs
        )
        if result.get("failed", False):
            raise AnsibleException(result)
        return result

    @need_ansible
    def get_variables(self):
        """Returns a dict of ansible variables

        >>> host.ansible.get_variables()
        {
            'inventory_hostname': 'localhost',
            'group_names': ['ungrouped'],
            'foo': 'bar',
        }

        """
        return self._host.backend.get_variables()

    def __repr__(self):
        return "<ansible>"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

from testinfra.modules.base import Module


class Docker(Module):
    """Test docker containers running on system.

    Example:

    >>> nginx = host.docker("app_nginx")
    >>> nginx.is_running
    True
    >>> nginx.id
    '7e67dc7495ca8f451d346b775890bdc0fb561ecdc97b68fb59ff2f77b509a8fe'
    >>> nginx.name
    'app_nginx'
    """

    def __init__(self, name):
        self._name = name
        super().__init__()

    def inspect(self):
        output = self.check_output("docker inspect %s", self._name)
        return json.loads(output)[0]

    @property
    def is_running(self):
        return self.inspect()["State"]["Running"]

    @property
    def is_restarting(self):
        return self.inspect()["State"]["Restarting"]

    @property
    def id(self):
        return self.inspect()["Id"]

    @property
    def name(self):
        return self.inspect()["Name"][1:]  # get rid of slash in front

    @classmethod
    def client_version(cls):
        """Docker client version"""
        return cls.version("{{.Client.Version}}")

    @classmethod
    def server_version(cls):
        """Docker server version"""
        return cls.version("{{.Server.Version}}")

    @classmethod
    def version(cls, format=None):
        """Docker version_ with an optional format (Go template).

        >>> host.docker.version()
        Client: Docker Engine - Community
        ...
        >>> host.docker.version("{{.Client.Context}}"))
        default

        .. _version: https://docs.docker.com/engine/reference/commandline/version/
        """
        cmd = "docker version"
        if format:
            cmd = f"{cmd} --format '{format}'"
        return cls.check_output(cmd)

    @classmethod
    def get_containers(cls, **filters):
        """Return a list of containers

        By default return list of all containers, including non-running
        containers.

        Filtering can be done using filters keys defined on
        https://docs.docker.com/engine/reference/commandline/ps/#filtering

        Multiple filters for a given key is handled by giving a list of string
        as value.

        >>> host.docker.get_containers()
        [<docker nginx>, <docker redis>, <docker app>]
        # Get all running containers
        >>> host.docker.get_containers(status="running")
        [<docker app>]
        # Get containers named "nginx"
        >>> host.docker.get_containers(name="nginx")
        [<docker nginx>]
        # Get containers named "nginx" or "redis"
        >>> host.docker.get_containers(name=["nginx", "redis"])
        [<docker nginx>, <docker redis>]
        """
        cmd = "docker ps --all --quiet --format '{{.Names}}'"
        args = []
        for key, value in filters.items():
            values = value if isinstance(value, (list, tuple)) else [value]
            for v in values:
                cmd += " --filter %s=%s"
                args += [key, v]
        result = []
        for docker_id in cls(None).check_output(cmd, *args).splitlines():
            result.append(cls(docker_id))
        return result

    def __repr__(self):
        return f"<docker {self._name}>"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os
from collections.abc import Iterable
from typing import Any

import testinfra.backend
import testinfra.backend.base
import testinfra.modules
import testinfra.modules.base


class Host:
    _host_cache: dict[tuple[str, frozenset[tuple[str, Any]]], "Host"] = {}
    _hosts_cache: dict[
        tuple[frozenset[str], frozenset[tuple[str, Any]]], list["Host"]
    ] = {}

    def __init__(self, backend: testinfra.backend.base.BaseBackend):
        self.backend = backend
        super().__init__()

    def __repr__(self) -> str:
        return f"<testinfra.host.Host {self.backend.get_pytest_id()}>"

    @functools.cached_property
    def has_command_v(self) -> bool:
        """Return True if `command -v` is available"""
        return self.run("command -v command").rc == 0

    def exists(self, command: str) -> bool:
        """Return True if given command exist in $PATH"""
        if self.has_command_v:
            out = self.run("command -v %s", command)
        else:
            out = self.run_expect([0, 1], "which %s", command)
        return out.rc == 0

    def find_command(
        self, command: str, extrapaths: Iterable[str] = ("/sbin", "/usr/sbin")
    ) -> str:
        """Return path of given command

        raise ValueError if command cannot be found
        """
        if self.has_command_v:
            out = self.run("command -v %s", command)
        else:
            out = self.run_expect([0, 1], "which %s", command)
        if out.rc == 0:
            return out.stdout.rstrip("\r\n")
        for basedir in extrapaths:
            path = os.path.join(basedir, command)
            if self.exists(path):
                return path
        raise ValueError(f'cannot find "{command}" command')

    def run(
        self, command: str, *args: str, **kwargs: Any
    ) -> testinfra.backend.base.CommandResult:
        """Run given command and return rc (exit status), stdout and stderr

        >>> cmd = host.run("ls -l /etc/passwd")
        >>> cmd.rc
        0
        >>> cmd.stdout
        '-rw-r--r-- 1 root root 1790 Feb 11 00:28 /etc/passwd\\n'
        >>> cmd.stderr
        ''
        >>> cmd.succeeded
        True
        >>> cmd.failed
        False


        Good practice: always use shell arguments quoting to avoid shell
        injection


        >>> cmd = host.run("ls -l -- %s", "/;echo inject")
        CommandResult(
            rc=2, stdout='',
            stderr=(
              'ls: cannot access /;echo inject: No such file or directory\\n'),
            command="ls -l '/;echo inject'")
        """
        return self.backend.run(command, *args, **kwargs)

    def run_expect(
        self, expected: list[int], command: str, *args: str, **kwargs: Any
    ) -> testinfra.backend.base.CommandResult:
        """Run command and check it return an expected exit status

        :param expected: A list of expected exit status
        :raises: AssertionError
        """
        __tracebackhide__ = True
        out = self.run(command, *args, **kwargs)
        assert out.rc in expected, f"Unexpected exit code {out.rc} for {out}"
        return out

    def run_test(
        self, command: str, *args: str, **kwargs: Any
    ) -> testinfra.backend.base.CommandResult:
        """Run command and check it return an exit status of 0 or 1

        :raises: AssertionError
        """
        return self.run_expect([0, 1], command, *args, **kwargs)

    def check_output(self, command: str, *args: str, **kwargs: Any) -> str:
        """Get stdout of a command which has run successfully

        :returns: stdout without trailing newline
        :raises: AssertionError
        """
        __tracebackhide__ = True
        out = self.run(command, *args, **kwargs)
        assert out.rc == 0, f"Unexpected exit code {out.rc} for {out}"
        return out.stdout.rstrip("\r\n")

    def __getattr__(self, name: str) -> type[testinfra.modules.base.Module]:
        if name in testinfra.modules.modules:
            module_class = testinfra.modules.get_module_class(name)
            obj = module_class.get_module(self)
            setattr(self, name, obj)
            return obj
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    @classmethod
    def get_host(cls, hostspec: str, **kwargs: Any) -> "Host":
        """Return a Host instance from `hostspec`

        `hostspec` should be like
        `<backend_type>://<name>?param1=value1&param2=value2`

        Params can also be passed in `**kwargs` (eg. get_host("local://",
        sudo=True) is equivalent to get_host("local://?sudo=true"))

        Examples::

        >>> get_host("local://", sudo=True)
        >>> get_host("paramiko://user@host", ssh_config="/path/my_ssh_config")
        >>> get_host("ansible://all?ansible_inventory=/etc/ansible/inventory")
        """
        key = (hostspec, frozenset(kwargs.items()))
        cache = cls._host_cache
        if key not in cache:
            backend = testinfra.backend.get_backend(hostspec, **kwargs)
            cache[key] = host = cls(backend)
            backend.set_host(host)
        return cache[key]

    @classmethod
    def get_hosts(cls, hosts: Iterable[str], **kwargs: Any) -> list["Host"]:
        key = (frozenset(hosts), frozenset(kwargs.items()))
        cache = cls._hosts_cache
        if key not in cache:
            cache[key] = []
            for backend in testinfra.backend.get_backends(hosts, **kwargs):
                host = cls(backend)
                backend.set_host(host)
                cache[key].append(host)
        return cache[key]


get_host = Host.get_host
get_hosts = Host.get_hosts



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import re

from testinfra.modules.base import InstanceModule


class SystemInfo(InstanceModule):
    """Return system information"""

    @functools.cached_property
    def sysinfo(self):
        sysinfo = {
            "type": None,
            "distribution": None,
            "codename": None,
            "release": None,
            "arch": None,
        }
        uname = self.run_expect([0, 1], "uname -s")
        if uname.rc == 1 or uname.stdout.lower().startswith("msys"):
            # FIXME: find a better way to detect windows here
            sysinfo.update(**self._get_windows_sysinfo())
            return sysinfo
        sysinfo["type"] = uname.stdout.rstrip("\r\n").lower()
        if sysinfo["type"] == "linux":
            sysinfo.update(**self._get_linux_sysinfo())
        elif sysinfo["type"] == "darwin":
            sysinfo.update(**self._get_darwin_sysinfo())
        else:
            # BSD
            sysinfo["release"] = self.check_output("uname -r")
            sysinfo["distribution"] = sysinfo["type"]
            sysinfo["codename"] = None

        sysinfo["arch"] = self.check_output("uname -m")
        return sysinfo

    def _get_linux_sysinfo(self):
        sysinfo = {}

        # https://www.freedesktop.org/software/systemd/man/os-release.html
        os_release = self.run("cat /etc/os-release")
        if os_release.rc == 0:
            for line in os_release.stdout.splitlines():
                for key, attname in (
                    ("ID=", "distribution"),
                    ("VERSION_ID=", "release"),
                    ("VERSION_CODENAME=", "codename"),
                ):
                    if line.startswith(key):
                        sysinfo[attname] = (
                            line[len(key) :].replace('"', "").replace("'", "").strip()
                        )
            # Arch doesn't have releases
            if "distribution" in sysinfo and sysinfo["distribution"] == "arch":
                sysinfo["release"] = "rolling"
            return sysinfo

        # RedHat / CentOS 6 haven't /etc/os-release
        redhat_release = self.run("cat /etc/redhat-release")
        if redhat_release.rc == 0:
            match = re.match(
                r"^(.+) release ([^ ]+) .*$", redhat_release.stdout.strip()
            )
            if match:
                sysinfo["distribution"], sysinfo["release"] = match.groups()
                return sysinfo

        # Alpine doesn't have /etc/os-release
        alpine_release = self.run("cat /etc/alpine-release")
        if alpine_release.rc == 0:
            sysinfo["distribution"] = "alpine"
            sysinfo["release"] = alpine_release.stdout.strip()
            return sysinfo

        # LSB
        lsb = self.run("lsb_release -a")
        if lsb.rc == 0:
            for line in lsb.stdout.splitlines():
                key, value = line.split(":", 1)
                key = key.strip().lower()
                value = value.strip().lower()
                if key == "distributor id":
                    sysinfo["distribution"] = value
                elif key == "release":
                    sysinfo["release"] = value
                elif key == "codename":
                    sysinfo["codename"] = value
            return sysinfo

        return sysinfo

    def _get_darwin_sysinfo(self):
        sysinfo = {}

        sw_vers = self.run("sw_vers")
        if sw_vers.rc == 0:
            for line in sw_vers.stdout.splitlines():
                key, value = line.split(":", 1)
                key = key.strip().lower()
                value = value.strip()
                if key == "productname":
                    sysinfo["distribution"] = value
                elif key == "productversion":
                    sysinfo["release"] = value

        return sysinfo

    def _get_windows_sysinfo(self):
        sysinfo = {}
        for line in self.check_output('systeminfo | findstr /B /C:"OS"').splitlines():
            key, value = line.split(":", 1)
            key = key.strip().replace(" ", "_").lower()
            value = value.strip()
            if key == "os_name":
                sysinfo["distribution"] = value
                sysinfo["type"] = value.split(" ")[1].lower()
            elif key == "os_version":
                sysinfo["release"] = value
        sysinfo["arch"] = self.check_output("echo %PROCESSOR_ARCHITECTURE%")
        return sysinfo

    @property
    def type(self):
        """OS type

        >>> host.system_info.type
        'linux'
        """
        return self.sysinfo["type"]

    @property
    def distribution(self):
        """Distribution name

        >>> host.system_info.distribution
        'debian'
        """
        return self.sysinfo["distribution"]

    @property
    def release(self):
        """Distribution release number

        >>> host.system_info.release
        '10.2'
        """
        return self.sysinfo["release"]

    @property
    def codename(self):
        """Release code name

        >>> host.system_info.codename
        'bullseye'
        """
        return self.sysinfo["codename"]

    @property
    def arch(self):
        """Host architecture

        >>> host.system_info.arch
        'x86_64'
        """
        return self.sysinfo["arch"]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import testinfra.modules.base

modules = {
    "addr": "addr:Addr",
    "ansible": "ansible:Ansible",
    "command": "command:Command",
    "docker": "docker:Docker",
    "podman": "podman:Podman",
    "environment": "environment:Environment",
    "file": "file:File",
    "group": "group:Group",
    "interface": "interface:Interface",
    "iptables": "iptables:Iptables",
    "mount_point": "mountpoint:MountPoint",
    "package": "package:Package",
    "pip": "pip:Pip",
    "process": "process:Process",
    "puppet_resource": "puppet:PuppetResource",
    "facter": "puppet:Facter",
    "salt": "salt:Salt",
    "service": "service:Service",
    "socket": "socket:Socket",
    "sudo": "sudo:Sudo",
    "supervisor": "supervisor:Supervisor",
    "sysctl": "sysctl:Sysctl",
    "system_info": "systeminfo:SystemInfo",
    "user": "user:User",
    "block_device": "blockdevice:BlockDevice",
}


def get_module_class(name: str) -> type["testinfra.modules.base.Module"]:
    modname, classname = modules[name].split(":")
    modname = ".".join([__name__, modname])
    module = importlib.import_module(modname)
    return getattr(module, classname)  # type: ignore[no-any-return]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Any, Optional

from testinfra.backend import base

try:
    import winrm
except ImportError:
    raise RuntimeError(
        "You must install the pywinrm package (pip install pywinrm) "
        "to use the winrm backend"
    ) from None

import winrm.protocol

_find_unsafe = re.compile(r"[^\w@%+=:,./-]", re.ASCII)


# (gtmanfred) This is copied from pipes.quote, but changed to use double quotes
# instead of single quotes.  This is used by the winrm backend.
def _quote(s: str) -> str:
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"
    if _find_unsafe.search(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return '"' + s.replace('"', '"\'"\'"') + '"'


class WinRMBackend(base.BaseBackend):
    """Run command through winrm command"""

    NAME = "winrm"

    def __init__(
        self,
        hostspec: str,
        no_ssl: bool = False,
        no_verify_ssl: bool = False,
        read_timeout_sec: Optional[int] = None,
        operation_timeout_sec: Optional[int] = None,
        *args: Any,
        **kwargs: Any,
    ):
        self.host = self.parse_hostspec(hostspec)
        self.conn_args: dict[str, Any] = {
            "endpoint": "{}://{}{}/wsman".format(
                "http" if no_ssl else "https",
                self.host.name,
                f":{self.host.port}" if self.host.port else "",
            ),
            "transport": "ntlm",
            "username": self.host.user,
            "password": self.host.password,
        }
        if no_verify_ssl:
            self.conn_args["server_cert_validation"] = "ignore"
        if read_timeout_sec is not None:
            self.conn_args["read_timeout_sec"] = read_timeout_sec
        if operation_timeout_sec is not None:
            self.conn_args["operation_timeout_sec"] = operation_timeout_sec
        super().__init__(self.host.name, *args, **kwargs)

    def run(self, command: str, *args: str, **kwargs: Any) -> base.CommandResult:
        return self.run_winrm(self.get_command(command, *args))

    def run_winrm(self, command: str) -> base.CommandResult:
        p = winrm.protocol.Protocol(**self.conn_args)
        shell_id = p.open_shell()
        command_id = p.run_command(shell_id, command)
        stdout, stderr, rc = p.get_command_output(shell_id, command_id)
        p.cleanup_command(shell_id, command_id)
        p.close_shell(shell_id)
        return self.result(rc, self.encode(command), stdout, stderr)

    @staticmethod
    def quote(command: str, *args: str) -> str:
        if args:
            return command % tuple(_quote(a) for a in args)
        return command



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import shutil
import sys
import tempfile
import time
from typing import AnyStr, cast

import pytest

import testinfra
import testinfra.host
import testinfra.modules


@pytest.fixture(scope="module")
def _testinfra_host(request: pytest.FixtureRequest) -> testinfra.host.Host:
    return cast(testinfra.host.Host, request.param)


@pytest.fixture(scope="module")
def host(_testinfra_host: testinfra.host.Host) -> testinfra.host.Host:
    return _testinfra_host


host.__doc__ = testinfra.host.Host.__doc__


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("testinfra")
    group.addoption(
        "--connection",
        action="store",
        dest="connection",
        help=(
            "Remote connection backend (paramiko, ssh, safe-ssh, "
            "salt, docker, ansible, podman)"
        ),
    )
    group.addoption(
        "--hosts",
        action="store",
        dest="hosts",
        help="Hosts list (comma separated)",
    )
    group.addoption(
        "--ssh-config",
        action="store",
        dest="ssh_config",
        help="SSH config file",
    )
    group.addoption(
        "--ssh-extra-args",
        action="store",
        dest="ssh_extra_args",
        help="SSH extra args",
    )
    group.addoption(
        "--ssh-identity-file",
        action="store",
        dest="ssh_identity_file",
        help="SSH identify file",
    )
    group.addoption(
        "--sudo",
        action="store_true",
        dest="sudo",
        help="Use sudo",
    )
    group.addoption(
        "--sudo-user",
        action="store",
        dest="sudo_user",
        help="sudo user",
    )
    group.addoption(
        "--ansible-inventory",
        action="store",
        dest="ansible_inventory",
        help="Ansible inventory file",
    )
    group.addoption(
        "--force-ansible",
        action="store_true",
        dest="force_ansible",
        help=(
            "Force use of ansible connection backend only (slower but all "
            "ansible connection options are handled)"
        ),
    )
    group.addoption(
        "--nagios",
        action="store_true",
        dest="nagios",
        help="Nagios plugin",
    )


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    if "_testinfra_host" in metafunc.fixturenames:
        if metafunc.config.option.hosts is not None:
            hosts = metafunc.config.option.hosts.split(",")
        elif hasattr(metafunc.module, "testinfra_hosts"):
            hosts = metafunc.module.testinfra_hosts
        else:
            hosts = [None]
        params = testinfra.get_hosts(
            hosts,
            connection=metafunc.config.option.connection,
            ssh_config=metafunc.config.option.ssh_config,
            ssh_identity_file=metafunc.config.option.ssh_identity_file,
            sudo=metafunc.config.option.sudo,
            sudo_user=metafunc.config.option.sudo_user,
            ansible_inventory=metafunc.config.option.ansible_inventory,
            force_ansible=metafunc.config.option.force_ansible,
        )
        params = sorted(params, key=lambda x: x.backend.get_pytest_id())
        ids = [e.backend.get_pytest_id() for e in params]
        metafunc.parametrize(
            "_testinfra_host", params, ids=ids, scope="module", indirect=True
        )


class NagiosReporter:
    def __init__(self, out):
        self.passed = 0
        self.failed = 0
        self.skipped = 0
        self.start_time = time.time()
        self.total_time = None
        self.out = out

    def pytest_runtest_logreport(self, report: pytest.TestReport) -> None:
        if report.passed:
            if report.when == "call":  # ignore setup/teardown
                self.passed += 1
        elif report.failed:
            self.failed += 1
        elif report.skipped:
            self.skipped += 1

    def report(self) -> int:
        if self.failed:
            status = b"CRITICAL"
            ret = 2
        else:
            status = b"OK"
            ret = 0

        out = sys.stdout.buffer
        out.write(
            (b"TESTINFRA %s - %d passed, %d failed, %d skipped in %.2f seconds\n")
            % (
                status,
                self.passed,
                self.failed,
                self.skipped,
                time.time() - self.start_time,
            )
        )
        self.out.seek(0)
        shutil.copyfileobj(self.out, out)
        return ret


class SpooledTemporaryFile(tempfile.SpooledTemporaryFile[AnyStr]):
    def __init__(self, *args, **kwargs):
        if "b" in kwargs.get("mode", "b"):
            self._out_encoding = kwargs.pop("encoding")
        else:
            self._out_encoding = kwargs.get("encoding")
        super().__init__(*args, **kwargs)

    def write(self, s):
        # avoid traceback in py.io.terminalwriter.write_out
        # TypeError: a bytes-like object is required, not 'str'
        if isinstance(s, str):
            s = s.encode(self._out_encoding)
        return super().write(s)


@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    if config.getoption("--verbose", 0) > 1:
        root = logging.getLogger()
        if not root.handlers:
            root.addHandler(logging.NullHandler())
        logging.getLogger("testinfra").setLevel(logging.DEBUG)
    if config.getoption("--nagios"):
        # disable & re-enable terminalreporter to write in a tempfile
        reporter = config.pluginmanager.getplugin("terminalreporter")
        if reporter:
            out = SpooledTemporaryFile(encoding=sys.stdout.encoding)
            config.pluginmanager.unregister(reporter)
            reporter = reporter.__class__(config, out)
            config.pluginmanager.register(reporter, "terminalreporter")
            config.pluginmanager.register(NagiosReporter(out), "nagiosreporter")


@pytest.hookimpl(trylast=True)
def pytest_sessionfinish(session, exitstatus):
    reporter = session.config.pluginmanager.getplugin("nagiosreporter")
    if reporter:
        session.exitstatus = reporter.report()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from testinfra.modules.base import Module


class File(Module):
    """Test various files attributes"""

    def __init__(self, path):
        self.path = path
        super().__init__()

    @property
    def exists(self):
        """Test if file exists

        >>> host.file("/etc/passwd").exists
        True
        >>> host.file("/nonexistent").exists
        False

        """
        return self.run_test("test -e %s", self.path).rc == 0

    @property
    def is_file(self):
        """Test if the path is a regular file"""
        return self.run_test("test -f %s", self.path).rc == 0

    @property
    def is_directory(self):
        """Test if the path exists and a directory"""
        return self.run_test("test -d %s", self.path).rc == 0

    @property
    def is_executable(self):
        """Test if the path exists and permission to execute is granted"""
        return self.run_test("test -x %s", self.path).rc == 0

    @property
    def is_pipe(self):
        """Test if the path exists and is a pipe"""
        return self.run_test("test -p %s", self.path).rc == 0

    @property
    def is_socket(self):
        """Test if the path exists and is a socket"""
        return self.run_test("test -S %s", self.path).rc == 0

    @property
    def is_symlink(self):
        """Test if the path exists and is a symbolic link"""
        return self.run_test("test -L %s", self.path).rc == 0

    @property
    def linked_to(self):
        """Resolve symlink

        >>> host.file("/var/lock").linked_to
        '/run/lock'
        """
        res = self.run_expect([0, 127], "realpath %s", self.path)
        if res.rc == 0:
            return res.stdout.strip()
        return self.check_output("readlink -f %s", self.path)

    @property
    def user(self):
        """Return file owner as string

        >>> host.file("/etc/passwd").user
        'root'
        """
        raise NotImplementedError

    @property
    def uid(self):
        """Return file user id as integer

        >>> host.file("/etc/passwd").uid
        0
        """
        raise NotImplementedError

    @property
    def group(self):
        """Return file group name as string"""
        raise NotImplementedError

    @property
    def gid(self):
        """Return file group id as integer"""
        raise NotImplementedError

    @property
    def mode(self):
        """Return file mode as octal integer

        >>> host.file("/etc/shadow").mode
        416  # Oo640 octal
        >>> host.file("/etc/shadow").mode == 0o640
        True
        >>> oct(host.file("/etc/shadow").mode) == '0o640'
        True

        You can also utilize the file mode constants from
        the stat_ library for testing file mode.

        >>> import stat
        >>> host.file("/etc/shadow").mode == stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP
        True

        .. _oct(x): https://docs.python.org/3/library/functions.html#oct
        .. _stat: https://docs.python.org/3/library/stat.html
        """
        raise NotImplementedError

    def contains(self, pattern):
        """Checks content of file for pattern

        This uses grep and thus follows the grep regex syntax.
        """
        return self.run_test("grep -qs -- %s %s", pattern, self.path).rc == 0

    @property
    def md5sum(self):
        """Compute the MD5 message digest of the file content"""
        raise NotImplementedError

    @property
    def sha256sum(self):
        """Compute the SHA256 message digest of the file content"""
        raise NotImplementedError

    def _get_content(self, decode):
        out = self.run_test("cat -- %s", self.path)
        if out.rc != 0:
            raise RuntimeError(f"Unexpected output {out}")
        if decode:
            return out.stdout
        return out.stdout_bytes

    @property
    def content(self):
        """Return file content as bytes

        >>> host.file("/tmp/foo").content
        b'caf\\xc3\\xa9'
        """
        return self._get_content(False)

    @property
    def content_string(self):
        """Return file content as string

        >>> host.file("/tmp/foo").content_string
        'café'
        """
        return self._get_content(True)

    @property
    def mtime(self):
        """Return time of last modification as datetime.datetime object

        >>> host.file("/etc/passwd").mtime
        datetime.datetime(2015, 3, 15, 20, 25, 40)
        """
        raise NotImplementedError

    @property
    def size(self):
        """Return size of file in bytes"""
        raise NotImplementedError

    def listdir(self):
        """Return list of items under the directory

        >>> host.file("/tmp").listdir()
        ['foo_file', 'bar_dir']
        """
        out = self.run_test("ls -1 -q -- %s", self.path)
        if out.rc != 0:
            raise RuntimeError(f"Unexpected output {out}")
        return out.stdout.splitlines()

    def __repr__(self):
        return f"<file {self.path}>"

    def __eq__(self, other):
        if isinstance(other, File):
            return self.path == other.path
        if isinstance(other, str):
            return self.path == other
        return False

    @classmethod
    def get_module_class(cls, host):
        if host.system_info.type == "linux":
            return GNUFile
        if host.system_info.type == "netbsd":
            return NetBSDFile
        if host.system_info.type.endswith("bsd"):
            return BSDFile
        if host.system_info.type == "darwin":
            return DarwinFile
        if host.system_info.type == "windows":
            return WindowsFile
        raise NotImplementedError


class GNUFile(File):
    @property
    def user(self):
        return self.check_output("stat -Lc %%U %s", self.path)

    @property
    def uid(self):
        return int(self.check_output("stat -Lc %%u %s", self.path))

    @property
    def group(self):
        return self.check_output("stat -Lc %%G %s", self.path)

    @property
    def gid(self):
        return int(self.check_output("stat -Lc %%g %s", self.path))

    @property
    def mode(self):
        # Supply a base of 8 when parsing an octal integer
        # e.g. int('644', 8) -> 420
        return int(self.check_output("stat -Lc %%a %s", self.path), 8)

    @property
    def mtime(self):
        ts = self.check_output("stat -Lc %%Y %s", self.path)
        return datetime.datetime.fromtimestamp(float(ts))

    @property
    def size(self):
        return int(self.check_output("stat -Lc %%s %s", self.path))

    @property
    def inode(self):
        return int(self.check_output("stat -Lc %%i %s", self.path))

    @property
    def md5sum(self):
        return self.check_output("md5sum %s | cut -d' ' -f1", self.path)

    @property
    def sha256sum(self):
        return self.check_output("sha256sum %s | cut -d ' ' -f 1", self.path)


class BSDFile(File):
    @property
    def user(self):
        return self.check_output("stat -f %%Su %s", self.path)

    @property
    def uid(self):
        return int(self.check_output("stat -f %%u %s", self.path))

    @property
    def group(self):
        return self.check_output("stat -f %%Sg %s", self.path)

    @property
    def gid(self):
        return int(self.check_output("stat -f %%g %s", self.path))

    @property
    def mode(self):
        # Supply a base of 8 when parsing an octal integer
        # e.g. int('644', 8) -> 420
        return int(self.check_output("stat -f %%Lp %s", self.path), 8)

    @property
    def mtime(self):
        ts = self.check_output("stat -f %%m %s", self.path)
        return datetime.datetime.fromtimestamp(float(ts))

    @property
    def size(self):
        return int(self.check_output("stat -f %%z %s", self.path))

    @property
    def md5sum(self):
        return self.check_output("md5 < %s", self.path)

    @property
    def sha256sum(self):
        return self.check_output("sha256 < %s", self.path)


class DarwinFile(BSDFile):
    @property
    def linked_to(self):
        link_script = f"""
        TARGET_FILE='{self.path}'
        cd `dirname $TARGET_FILE`
        TARGET_FILE=`basename $TARGET_FILE`
        while [ -L "$TARGET_FILE" ]
        do
            TARGET_FILE=`readlink $TARGET_FILE`
            cd `dirname $TARGET_FILE`
            TARGET_FILE=`basename $TARGET_FILE`
        done
        PHYS_DIR=`pwd -P`
        RESULT=$PHYS_DIR/$TARGET_FILE
        echo $RESULT
        """
        return self.check_output(link_script)


class NetBSDFile(BSDFile):
    @property
    def sha256sum(self):
        return self.check_output("cksum -a sha256 < %s", self.path)


class WindowsFile(File):
    @property
    def exists(self):
        """Test if file exists

        >>> host.file(r"C:/Users").exists
        True
        >>> host.file(r"C:/nonexistent").exists
        False
        """

        return (
            self.check_output(r"powershell -command \"Test-Path '%s'\"", self.path)
            == "True"
        )

    @property
    def is_file(self):
        return (
            self.check_output(
                r"powershell -command \"(Get-Item '%s') -is [System.IO.FileInfo]\"",
                self.path,
            )
            == "True"
        )

    @property
    def is_directory(self):
        return (
            self.check_output(
                r"powershell -command \"(Get-Item '%s') -is [System.IO.DirectoryInfo]\"",
                self.path,
            )
            == "True"
        )

    @property
    def is_pipe(self):
        raise NotImplementedError

    @property
    def is_socket(self):
        raise NotImplementedError

    @property
    def is_symlink(self):
        return (
            self.check_output(
                r"powershell -command \"(Get-Item -Path '%s').Attributes -band [System.IO.FileAttributes]::ReparsePoint\"",
                self.path,
            )
            == "True"
        )

    @property
    def linked_to(self):
        """Resolve symlink

        >>> host.file("C:/Users/lock").linked_to
        'C:/Program Files/lock'
        """
        return self.check_output(
            r"powershell -command \"(Get-Item -Path '%s' -ReadOnly).FullName\"",
            self.path,
        )

    @property
    def user(self):
        raise NotImplementedError

    @property
    def uid(self):
        raise NotImplementedError

    @property
    def group(self):
        raise NotImplementedError

    @property
    def gid(self):
        raise NotImplementedError

    @property
    def mode(self):
        raise NotImplementedError

    def contains(self, pattern):
        """Checks content of file for pattern

        This follows the regex syntax.
        """
        return (
            self.run_test(
                r"powershell -command \"Select-String -Path '%s' -Pattern '%s'\"",
                self.path,
                pattern,
            ).stdout
            != ""
        )

    @property
    def md5sum(self):
        raise NotImplementedError

    @property
    def sha256sum(self):
        raise NotImplementedError

    def _get_content(self, decode):
        out = self.run_expect([0], r"powershell -command \"cat -- '%s'\"", self.path)
        if decode:
            return out.stdout
        return out.stdout_bytes

    @property
    def content(self):
        """Return file content as bytes

        >>> host.file("C:/Windows/Temp/foo").content
        b'caf\\xc3\\xa9'
        """
        return self._get_content(False)

    @property
    def content_string(self):
        """Return file content as string

        >>> host.file("C:/Windows/Temp/foo").content_string
        'café'
        """
        return self._get_content(True)

    @property
    def mtime(self):
        """Return time of last modification as datetime.datetime object

        >>> host.file("C:/Windows/passwd").mtime
        datetime.datetime(2015, 3, 15, 20, 25, 40)
        """
        date_time_str = self.check_output(
            r"powershell -command \"Get-ChildItem -Path '%s' | Select-Object -ExpandProperty LastWriteTime\"",
            self.path,
        )
        return datetime.datetime.strptime(
            date_time_str.strip(), "%A, %B %d, %Y %I:%M:%S %p"
        )

    @property
    def size(self):
        """Return size of file in bytes"""
        return int(
            self.check_output(
                r"powershell -command \"Get-Item -Path '%s' | Select-Object -ExpandProperty Length\"",
                self.path,
            )
        )

    def listdir(self):
        """Return list of items under the directory

        >>> host.file("C:/Windows/Temp").listdir()
        ['foo_file', 'bar_dir']
        """
        out = self.check_output(
            r"powershell -command \"Get-ChildItem -Path '%s' | Select-Object -ExpandProperty Name\"",
            self.path,
        )
        return [item.strip() for item in out.strip().split("\n")]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import os
import urllib.parse
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import testinfra.backend.base

BACKENDS = {
    "local": "testinfra.backend.local.LocalBackend",
    "ssh": "testinfra.backend.ssh.SshBackend",
    "safe-ssh": "testinfra.backend.ssh.SafeSshBackend",
    "paramiko": "testinfra.backend.paramiko.ParamikoBackend",
    "salt": "testinfra.backend.salt.SaltBackend",
    "docker": "testinfra.backend.docker.DockerBackend",
    "podman": "testinfra.backend.podman.PodmanBackend",
    "ansible": "testinfra.backend.ansible.AnsibleBackend",
    "kubectl": "testinfra.backend.kubectl.KubectlBackend",
    "winrm": "testinfra.backend.winrm.WinRMBackend",
    "lxc": "testinfra.backend.lxc.LxcBackend",
    "openshift": "testinfra.backend.openshift.OpenShiftBackend",
    "chroot": "testinfra.backend.chroot.ChrootBackend",
}


def get_backend_class(connection: str) -> type["testinfra.backend.base.BaseBackend"]:
    try:
        classpath = BACKENDS[connection]
    except KeyError:
        raise RuntimeError(f"Unknown connection type '{connection}'") from None
    module, name = classpath.rsplit(".", 1)
    return getattr(importlib.import_module(module), name)  # type: ignore[no-any-return]


def parse_hostspec(hostspec: str) -> tuple[str, dict[str, Any]]:
    kw: dict[str, Any] = {}
    if hostspec is not None and "://" in hostspec:
        url = urllib.parse.urlparse(hostspec)
        kw["connection"] = url.scheme
        host = url.netloc
        query = urllib.parse.parse_qs(url.query)
        for key in ("sudo", "ssl", "no_ssl", "no_verify_ssl", "force_ansible"):
            if query.get(key, ["false"])[0].lower() == "true":
                kw[key] = True
        for key in (
            "sudo_user",
            "namespace",
            "container",
            "read_timeout_sec",
            "operation_timeout_sec",
            "timeout",
            "controlpersist",
            "kubeconfig",
            "context",
        ):
            if key in query:
                kw[key] = query[key][0]
        for key in (
            "ssh_config",
            "ansible_inventory",
            "ssh_identity_file",
        ):
            if key in query:
                kw[key] = os.path.expanduser(query[key][0])
    else:
        host = hostspec
    return host, kw


def get_backend(hostspec: str, **kwargs: Any) -> "testinfra.backend.base.BaseBackend":
    host, kw = parse_hostspec(hostspec)
    for k, v in kwargs.items():
        kw.setdefault(k, v)
    kw.setdefault("connection", "paramiko")
    klass = get_backend_class(kw["connection"])
    if kw["connection"] == "local":
        return klass(**kw)
    return klass(host, **kw)


def get_backends(
    hosts: Iterable[str], **kwargs: Any
) -> list["testinfra.backend.base.BaseBackend"]:
    backends = {}
    for hostspec in hosts:
        host, kw = parse_hostspec(hostspec)
        for k, v in kwargs.items():
            kw.setdefault(k, v)
        connection = kw.get("connection")
        if host is None and connection is None:
            connection = "local"
        elif host is not None and connection is None:
            connection = "paramiko"
        klass = get_backend_class(connection)
        for name in klass.get_hosts(host, **kw):
            key = (name, frozenset(kw.items()))
            if key in backends:
                continue
            backend = klass(**kw) if connection == "local" else klass(name, **kw)
            backends[key] = backend
    return list(backends.values())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import json
import re

from testinfra.modules.base import Module


class Interface(Module):
    """Test network interfaces

    >>> host.interface("eth0").exists
    True

    Optionally, the protocol family to use can be enforced.

    >>> host.interface("eth0", "inet6").addresses
    ['fe80::e291:f5ff:fe98:6b8c']
    """

    def __init__(self, name, family=None):
        self.name = name
        self.family = family
        super().__init__()

    @property
    def exists(self):
        raise NotImplementedError

    @property
    def speed(self):
        raise NotImplementedError

    @property
    def addresses(self):
        """Return ipv4 and ipv6 addresses on the interface

        >>> host.interface("eth0").addresses
        ['192.168.31.254', '192.168.31.252', 'fe80::e291:f5ff:fe98:6b8c']
        """
        raise NotImplementedError

    @property
    def link(self):
        """Return the link properties associated with the interface.

        >>> host.interface("lo").link
        {'address': '00:00:00:00:00:00',
        'broadcast': '00:00:00:00:00:00',
        'flags': ['LOOPBACK', 'UP', 'LOWER_UP'],
        'group': 'default',
        'ifindex': 1,
        'ifname': 'lo',
        'link_type': 'loopback',
        'linkmode': 'DEFAULT',
        'mtu': 65536,
        'operstate': 'UNKNOWN',
        'qdisc': 'noqueue',
        'txqlen': 1000}
        """
        raise NotImplementedError

    def routes(self, scope=None):
        """Return the routes associated with the interface, optionally filtered by scope
        ("host", "link" or "global").

        >>> host.interface("eth0").routes()
        [{'dst': 'default',
        'flags': [],
        'gateway': '192.0.2.1',
        'metric': 3003,
        'prefsrc': '192.0.2.5',
        'protocol': 'dhcp'},
        {'dst': '192.0.2.0/24',
        'flags': [],
        'metric': 3003,
        'prefsrc': '192.0.2.5',
        'protocol': 'dhcp',
        'scope': 'link'}]
        """
        raise NotImplementedError

    def __repr__(self):
        return f"<interface {self.name}>"

    @classmethod
    def get_module_class(cls, host):
        if host.system_info.type == "linux":
            return LinuxInterface
        if host.system_info.type.endswith("bsd"):
            return BSDInterface
        raise NotImplementedError

    @classmethod
    def names(cls):
        """Return the names of all the interfaces.

        >>> host.interface.names()
        ['lo', 'tunl0', 'ip6tnl0', 'eth0']
        """
        raise NotImplementedError

    @classmethod
    def default(cls, family=None):
        """Return the interface used for the default route.

        >>> host.interface.default()
        <interface eth0>

        Optionally, the protocol family to use can be enforced.

        >>> host.interface.default("inet6")
        None
        """
        raise NotImplementedError


class LinuxInterface(Interface):
    @functools.cached_property
    def _ip(self):
        ip_cmd = self.find_command("ip")
        if self.family is not None:
            ip_cmd = f"{ip_cmd} -f {self.family}"
        return ip_cmd

    @property
    def exists(self):
        return self.run_test(f"{self._ip} link show %s", self.name).rc == 0

    @property
    def speed(self):
        return int(self.check_output("cat /sys/class/net/%s/speed", self.name))

    @property
    def addresses(self):
        stdout = self.check_output(f"{self._ip} addr show %s", self.name)
        addrs = []
        for line in stdout.splitlines():
            splitted = [e.strip() for e in line.split(" ") if e]
            if splitted and splitted[0] in ("inet", "inet6"):
                addrs.append(splitted[1].split("/", 1)[0])
        return addrs

    @property
    def link(self):
        return json.loads(
            self.check_output(f"{self._ip} --json link show %s", self.name)
        )

    def routes(self, scope=None):
        cmd = f"{self._ip} --json route list dev %s"

        if scope is None:
            out = self.check_output(cmd, self.name)
        else:
            out = self.check_output(cmd + " scope %s", self.name, scope)

        return json.loads(out)

    @classmethod
    def default(cls, family=None):
        _default = cls(None, family=family)
        out = cls.check_output(f"{_default._ip} route ls")
        for line in out.splitlines():
            if "default" in line:
                match = re.search(r"dev\s(\S+)", line)
                if match:
                    _default.name = match.group(1)
        return _default

    @classmethod
    def names(cls):
        # -o is to tell the ip command to return 1 line per interface
        out = cls.check_output(f"{cls(None)._ip} -o link show")
        interfaces = []
        for line in out.splitlines():
            interfaces.append(line.strip().split(": ", 2)[1].split("@", 1)[0])
        return interfaces


class BSDInterface(Interface):
    @property
    def exists(self):
        return self.run_test("ifconfig %s", self.name).rc == 0

    @property
    def speed(self):
        raise NotImplementedError

    @property
    def addresses(self):
        stdout = self.check_output("ifconfig %s", self.name)
        addrs = []
        for line in stdout.splitlines():
            if line.startswith("\tinet "):
                addrs.append(line.split(" ", 2)[1])
            elif line.startswith("\tinet6 "):
                addr = line.split(" ", 2)[1]
                if "%" in addr:
                    addr = addr.split("%", 1)[0]
                addrs.append(addr)
        return addrs



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from testinfra.modules.base import Module


class BlockDevice(Module):
    """Information for block device.

    Should be used with sudo or under root.

    If device is not a block device, RuntimeError is raised.
    """

    @property
    def _data(self):
        raise NotImplementedError

    def __init__(self, device):
        self.device = device
        super().__init__()

    @property
    def is_partition(self):
        """Return True if the device is a partition.

        >>> host.block_device("/dev/sda1").is_partition
        True

        >>> host.block_device("/dev/sda").is_partition
        False


        """
        return self._data["start_sector"] > 0

    @property
    def size(self):
        """Return size if the device in bytes.

        >>> host.block_device("/dev/sda1").size
        512110190592

        """
        return self._data["size"]

    @property
    def sector_size(self):
        """Return sector size for the device in bytes.

        >>> host.block_device("/dev/sda1").sector_size
        512
        """
        return self._data["sector_size"]

    @property
    def block_size(self):
        """Return block size for the device in bytes.

        >>> host.block_device("/dev/sda").block_size
        4096
        """
        return self._data["block_size"]

    @property
    def start_sector(self):
        """Return start sector of the device on the underlying device.

           Usually the value is zero for full devices and is non-zero
           for partitions.

        >>> host.block_device("/dev/sda1").start_sector
        2048

        >>> host.block_device("/dev/md0").start_sector
        0
        """
        return self._data["sector_size"]

    @property
    def is_writable(self):
        """Return True if device is writable (have no RO status)

        >>> host.block_device("/dev/sda").is_writable
        True

        >>> host.block_device("/dev/loop1").is_writable
        False
        """
        mode = self._data["rw_mode"]
        if mode == "rw":
            return True
        if mode == "ro":
            return False
        raise ValueError(f"Unexpected value for rw: {mode}")

    @property
    def ra(self):
        """Return Read Ahead for the device in 512-bytes sectors.

        >>> host.block_device("/dev/sda").ra
        256
        """
        return self._data["read_ahead"]

    @classmethod
    def get_module_class(cls, host):
        if host.system_info.type == "linux":
            return LinuxBlockDevice
        raise NotImplementedError

    def __repr__(self):
        return f"<BlockDevice(path={self.device})>"


class LinuxBlockDevice(BlockDevice):
    @functools.cached_property
    def _data(self):
        header = ["RO", "RA", "SSZ", "BSZ", "StartSec", "Size", "Device"]
        command = "blockdev  --report %s"
        blockdev = self.run(command, self.device)
        if blockdev.rc != 0:
            raise RuntimeError(f"Failed to gather data: {blockdev.stderr}")
        output = blockdev.stdout.splitlines()
        if len(output) < 2:
            raise RuntimeError(f"No data from {self.device}")
        if output[0].split() != header:
            raise RuntimeError(f"Unknown output of blockdev: {output[0]}")
        fields = output[1].split()
        return {
            "rw_mode": str(fields[0]),
            "read_ahead": int(fields[1]),
            "sector_size": int(fields[2]),
            "block_size": int(fields[3]),
            "start_sector": int(fields[4]),
            "size": int(fields[5]),
        }



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testinfra.modules.base import InstanceModule


class Iptables(InstanceModule):
    """Test iptables rule exists"""

    def __init__(self):
        super().__init__()
        # support for -w argument (since 1.6.0)
        # https://git.netfilter.org/iptables/commit/?id=aaa4ace72b
        # centos 6 has no support
        # centos 7 has 1.4 patched
        self._has_w_argument = None

    def _iptables_command(self, version):
        if version == 4:
            iptables = "iptables"
        elif version == 6:
            iptables = "ip6tables"
        else:
            raise RuntimeError(f"Invalid version: {version}")
        if self._has_w_argument is False:
            return iptables
        else:
            return f"{iptables} -w 90"

    def _run_iptables(self, version, cmd, *args):
        ipt_cmd = f"{self._iptables_command(version)} {cmd}"
        if self._has_w_argument is None:
            result = self.run_expect([0, 2], ipt_cmd, *args)
            if result.rc == 2:
                self._has_w_argument = False
                return self._run_iptables(version, cmd, *args)
            else:
                self._has_w_argument = True
                return result.stdout.rstrip("\r\n")
        else:
            return self.check_output(ipt_cmd, *args)

    def rules(self, table="filter", chain=None, version=4):
        """Returns list of iptables rules

           Based on output of `iptables -t TABLE -S CHAIN` command

             optionally takes takes the following arguments:
               - table: defaults to `filter`
               - chain: defaults to all chains
               - version: default 4 (iptables), optionally 6 (ip6tables)

        >>> host.iptables.rules()
        [
            '-P INPUT ACCEPT',
            '-P FORWARD ACCEPT',
            '-P OUTPUT ACCEPT',
            '-A INPUT -i lo -j ACCEPT',
            '-A INPUT -j REJECT'
            '-A FORWARD -j REJECT'
        ]
        >>> host.iptables.rules("nat", "INPUT")
        ['-P PREROUTING ACCEPT']

        """
        cmd, args = "-t %s -S", [table]
        if chain:
            cmd += " %s"
            args += [chain]

        rules = []
        for line in self._run_iptables(version, cmd, *args).splitlines():
            line = line.replace("\t", " ")
            rules.append(line)
        return rules



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from typing import Any, Optional

from testinfra.backend import base


class SshBackend(base.BaseBackend):
    """Run command through ssh command"""

    NAME = "ssh"

    def __init__(
        self,
        hostspec: str,
        ssh_config: Optional[str] = None,
        ssh_identity_file: Optional[str] = None,
        timeout: int = 10,
        controlpath: Optional[str] = None,
        controlpersist: int = 60,
        ssh_extra_args: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ):
        self.host = self.parse_hostspec(hostspec)
        self.ssh_config = ssh_config
        self.ssh_identity_file = ssh_identity_file
        self.timeout = int(timeout)
        self.controlpath = controlpath
        self.controlpersist = int(controlpersist)
        self.ssh_extra_args = ssh_extra_args
        super().__init__(self.host.name, *args, **kwargs)

    def run(self, command: str, *args: str, **kwargs: Any) -> base.CommandResult:
        return self.run_ssh(self.get_command(command, *args))

    def _build_ssh_command(self, command: str) -> tuple[list[str], list[str]]:
        if not self.host.password:
            cmd = ["ssh"]
            cmd_args = []
        else:
            cmd = ["sshpass", "-p", "%s", "ssh"]
            cmd_args = [self.host.password]

        if self.ssh_extra_args:
            cmd.append(self.ssh_extra_args.replace("%", "%%"))
        if self.ssh_config:
            cmd.append("-F %s")
            cmd_args.append(self.ssh_config)
        if self.host.user:
            cmd.append("-o User=%s")
            cmd_args.append(self.host.user)
        if self.host.port:
            cmd.append("-o Port=%s")
            cmd_args.append(self.host.port)
        if self.ssh_identity_file:
            cmd.append("-i %s")
            cmd_args.append(self.ssh_identity_file)
        if "connecttimeout" not in (self.ssh_extra_args or "").lower():
            cmd.append(f"-o ConnectTimeout={self.timeout}")
        if self.controlpersist and (
            "controlmaster" not in (self.ssh_extra_args or "").lower()
        ):
            cmd.append(
                f"-o ControlMaster=auto -o ControlPersist={self.controlpersist}s"
            )
        if (
            "ControlMaster" in " ".join(cmd)
            and self.controlpath
            and ("controlpath" not in (self.ssh_extra_args or "").lower())
        ):
            cmd.append(f"-o ControlPath={self.controlpath}")
        cmd.append("%s %s")
        cmd_args.extend([self.host.name, command])
        return cmd, cmd_args

    def run_ssh(self, command: str) -> base.CommandResult:
        cmd, cmd_args = self._build_ssh_command(command)
        out = self.run_local(" ".join(cmd), *cmd_args)
        out.command = self.encode(command)
        if out.rc == 255:
            # ssh exits with the exit status of the remote command or with 255
            # if an error occurred.
            raise RuntimeError(out)
        return out


class SafeSshBackend(SshBackend):
    """Run command using ssh command but try to get a more sane output

    When using ssh (or a potentially bugged wrapper) additional output can be
    added in stdout/stderr and exit status may not be propagate correctly

    To avoid that kind of bugs, we wrap the command to have an output like
    this:

    TESTINFRA_START;EXIT_STATUS;STDOUT;STDERR;TESTINFRA_END

    where STDOUT/STDERR are base64 encoded, then we parse that magic string to
    get sanes variables
    """

    NAME = "safe-ssh"

    def run(self, command: str, *args: str, **kwargs: Any) -> base.CommandResult:
        orig_command = self.get_command(command, *args)
        orig_command = self.get_command("sh -c %s", orig_command)

        out = self.run_ssh(
            f"""of=$(mktemp)&&ef=$(mktemp)&&{orig_command} >$of 2>$ef; r=$?;"""
            """echo "TESTINFRA_START;$r;$(base64 < $of);$(base64 < $ef);"""
            """TESTINFRA_END";rm -f $of $ef"""
        )

        start = out.stdout.find("TESTINFRA_START;") + len("TESTINFRA_START;")
        end = out.stdout.find("TESTINFRA_END") - 1
        rc, stdout, stderr = out.stdout[start:end].split(";")
        return self.result(
            int(rc),
            self.encode(orig_command),
            base64.b64decode(stdout),
            base64.b64decode(stderr),
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from testinfra.modules.base import InstanceModule


class Salt(InstanceModule):
    """Run salt module functions


    >>> host.salt("pkg.version", "nginx")
    '1.6.2-5'
    >>> host.salt("pkg.version", ["nginx", "php5-fpm"])
    {'nginx': '1.6.2-5', 'php5-fpm': '5.6.7+dfsg-1'}
    >>> host.salt("grains.item", ["osarch", "mem_total", "num_cpus"])
    {'osarch': 'amd64', 'num_cpus': 4, 'mem_total': 15520}

    Run ``salt-call sys.doc`` to get a complete list of functions
    """

    def __call__(self, function, args=None, local=False, config=None):
        args = args or []
        if isinstance(args, str):
            args = [args]
        if self._host.backend.HAS_RUN_SALT:
            return self._host.backend.run_salt(function, args)
        cmd_args = []
        cmd = "salt-call --out=json"
        if local:
            cmd += " --local"
        if config is not None:
            cmd += " -c %s"
            cmd_args.append(config)
        cmd += " %s" + len(args) * " %s"
        cmd_args += [function] + args
        return json.loads(self.check_output(cmd, *cmd_args))["local"]

    def __repr__(self):
        return "<salt>"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import fnmatch
import functools
import ipaddress
import json
import os
import tempfile
from collections.abc import Iterator
from typing import Any, Callable, Optional, Union

import testinfra
import testinfra.host

__all__ = ["AnsibleRunner"]

local = testinfra.get_host("local://")


def get_ansible_config() -> configparser.ConfigParser:
    fname = os.environ.get("ANSIBLE_CONFIG")
    if not fname:
        for possible in (
            "ansible.cfg",
            os.path.join(os.path.expanduser("~"), ".ansible.cfg"),
            os.path.join("/", "etc", "ansible", "ansible.cfg"),
        ):
            if os.path.exists(possible):
                fname = possible
                break
    config = configparser.ConfigParser()
    if not fname:
        return config
    config.read(fname)
    return config


Inventory = dict[str, Any]


def get_ansible_inventory(
    config: configparser.ConfigParser, inventory_file: Optional[str]
) -> Inventory:
    # Disable ansible verbosity to avoid
    # https://github.com/ansible/ansible/issues/59973
    cmd = "ANSIBLE_VERBOSITY=0 ansible-inventory --list"
    args = []
    if inventory_file:
        cmd += " -i %s"
        args += [inventory_file]
    return json.loads(local.check_output(cmd, *args))  # type: ignore[no-any-return]


def get_ansible_host(
    config: configparser.ConfigParser,
    inventory: Inventory,
    host: str,
    ssh_config: Optional[str] = None,
    ssh_identity_file: Optional[str] = None,
) -> Optional[testinfra.host.Host]:
    if is_empty_inventory(inventory):
        if host == "localhost":
            return testinfra.get_host("local://")
        return None
    hostvars = inventory["_meta"].get("hostvars", {}).get(host, {})
    connection = hostvars.get("ansible_connection", "ssh")
    if connection not in (
        "smart",
        "ssh",
        "paramiko_ssh",
        "local",
        "docker",
        "community.docker.docker",
        "lxc",
        "lxd",
    ):
        # unhandled connection type, must use force_ansible=True
        return None
    connection = {
        "community.docker.docker": "docker",
        "lxd": "lxc",
        "paramiko_ssh": "paramiko",
        "smart": "ssh",
    }.get(connection, connection)

    options: dict[str, Any] = {
        "ansible_become": {
            "ini": {
                "section": "privilege_escalation",
                "key": "become",
            },
            "environment": "ANSIBLE_BECOME",
        },
        "ansible_become_user": {
            "ini": {
                "section": "privilege_escalation",
                "key": "become_user",
            },
            "environment": "ANSIBLE_BECOME_USER",
        },
        "ansible_port": {
            "ini": {
                "section": "defaults",
                "key": "remote_port",
            },
            "environment": "ANSIBLE_REMOTE_PORT",
        },
        "ansible_ssh_common_args": {
            "ini": {
                "section": "ssh_connection",
                "key": "ssh_common_args",
            },
            "environment": "ANSIBLE_SSH_COMMON_ARGS",
        },
        "ansible_ssh_extra_args": {
            "ini": {
                "section": "ssh_connection",
                "key": "ssh_extra_args",
            },
            "environment": "ANSIBLE_SSH_EXTRA_ARGS",
        },
        "ansible_user": {
            "ini": {
                "section": "defaults",
                "key": "remote_user",
            },
            "environment": "ANSIBLE_REMOTE_USER",
        },
    }

    def get_config(
        name: str, default: Union[None, bool, str] = None
    ) -> Union[None, bool, str]:
        value = default
        option = options.get(name, {})

        ini = option.get("ini")
        if ini:
            value = config.get(ini["section"], ini["key"], fallback=default)

        if name in hostvars:
            value = hostvars[name]

        var = option.get("environment")
        if var and var in os.environ:
            value = os.environ[var]

        return value

    testinfra_host = get_config("ansible_host", host)
    assert isinstance(testinfra_host, str), testinfra_host
    user = get_config("ansible_user")
    password = get_config("ansible_ssh_pass")
    port = get_config("ansible_port")

    kwargs: dict[str, Union[None, str, bool]] = {}
    if get_config("ansible_become", False):
        kwargs["sudo"] = True
    kwargs["sudo_user"] = get_config("ansible_become_user")
    if ssh_config is not None:
        kwargs["ssh_config"] = ssh_config
    if ssh_identity_file is not None:
        kwargs["ssh_identity_file"] = ssh_identity_file

    # Support both keys as advertised by Ansible
    if "ansible_ssh_private_key_file" in hostvars:
        kwargs["ssh_identity_file"] = hostvars["ansible_ssh_private_key_file"]
    elif "ansible_private_key_file" in hostvars:
        kwargs["ssh_identity_file"] = hostvars["ansible_private_key_file"]
    kwargs["ssh_extra_args"] = " ".join(
        [
            config.get("ssh_connection", "ssh_args", fallback=""),
            get_config("ansible_ssh_common_args", ""),  # type: ignore[list-item]
            get_config("ansible_ssh_extra_args", ""),  # type: ignore[list-item]
        ]
    ).strip()

    control_path = config.get("ssh_connection", "control_path", fallback="", raw=True)
    if control_path:
        control_path_dir = config.get(
            "persistent_connection", "control_path_dir", fallback="~/.ansible/cp"
        )
        control_path_dir = os.path.expanduser(control_path_dir)
        control_path_dir = os.path.normpath(control_path_dir)

        if os.path.isdir(control_path_dir):
            control_path = control_path % ({"directory": control_path_dir})
            control_path = control_path.replace("%", "%%")  # restore original "%%"
            kwargs["controlpath"] = control_path

    spec = f"{connection}://"

    # Fallback to user:password auth when identity file is not used
    if user and password and not kwargs.get("ssh_identity_file"):
        spec += f"{user}:{password}@"
    elif user:
        spec += f"{user}@"

    try:
        version = ipaddress.ip_address(testinfra_host).version
    except ValueError:
        version = None
    if version == 6:
        spec += "[" + testinfra_host + "]"
    else:
        spec += testinfra_host
    if port:
        spec += f":{port}"
    return testinfra.get_host(spec, **kwargs)


def itergroup(inventory: Inventory, group: str) -> Iterator[str]:
    for host in inventory.get(group, {}).get("hosts", []):
        yield host
    for g in inventory.get(group, {}).get("children", []):
        for host in itergroup(inventory, g):
            yield host


def is_empty_inventory(inventory: Inventory) -> bool:
    return not any(True for _ in itergroup(inventory, "all"))


class AnsibleRunner:
    _runners: dict[Optional[str], "AnsibleRunner"] = {}
    _known_options = {
        # Boolean arguments.
        "become": {
            "cli": "--become",
            "type": "boolean",
        },
        "check": {
            "cli": "--check",
            "type": "boolean",
        },
        "diff": {
            "cli": "--diff",
            "type": "boolean",
        },
        "one_line": {
            "cli": "--one-line",
            "type": "boolean",
        },
        # String arguments.
        "become_method": {
            "cli": "--become-method",
            "type": "string",
        },
        "become_user": {
            "cli": "--become-user",
            "type": "string",
        },
        "user": {
            "cli": "--user",
            "type": "string",
        },
        # Arguments serialized as JSON.
        "extra_vars": {
            "cli": "--extra-vars",
            "type": "json",
        },
    }

    def __init__(self, inventory_file: Optional[str] = None):
        self.inventory_file = inventory_file
        self._host_cache: dict[str, Optional[testinfra.host.Host]] = {}
        super().__init__()

    def get_hosts(self, pattern: str = "all") -> list[str]:
        inventory = self.inventory
        result = set()
        if is_empty_inventory(inventory):
            # empty inventory should not return any hosts except for localhost
            if pattern == "localhost":
                result.add("localhost")
            else:
                raise RuntimeError(
                    "No inventory was parsed (missing file ?), "
                    "only implicit localhost is available"
                )
        else:
            for group in inventory:
                groupmatch = fnmatch.fnmatch(group, pattern)
                if groupmatch:
                    result |= set(itergroup(inventory, group))
                for host in inventory[group].get("hosts", []):
                    if fnmatch.fnmatch(host, pattern):
                        result.add(host)
        return sorted(result)

    @functools.cached_property
    def inventory(self) -> Inventory:
        return get_ansible_inventory(self.ansible_config, self.inventory_file)

    @functools.cached_property
    def ansible_config(self) -> configparser.ConfigParser:
        return get_ansible_config()

    def get_variables(self, host: str) -> dict[str, Any]:
        inventory = self.inventory
        # inventory_hostname, group_names and groups are for backward
        # compatibility with testinfra 2.X
        hostvars: dict[str, Any] = inventory["_meta"].get("hostvars", {}).get(host, {})
        hostvars.setdefault("inventory_hostname", host)
        group_names = []
        groups = {}

        for group in sorted(inventory):
            if group == "_meta":
                continue
            groups[group] = sorted(itergroup(inventory, group))
            if host in groups[group]:
                group_names.append(group)

        hostvars.setdefault("group_names", group_names)
        hostvars.setdefault("groups", groups)
        return hostvars

    def get_host(self, host: str, **kwargs: Any) -> Optional[testinfra.host.Host]:
        try:
            return self._host_cache[host]
        except KeyError:
            self._host_cache[host] = get_ansible_host(
                self.ansible_config, self.inventory, host, **kwargs
            )
            return self._host_cache[host]

    def options_to_cli(self, options: dict[str, Any]) -> tuple[str, list[str]]:
        verbose = options.pop("verbose", 0)

        args = {"become": False, "check": True}
        args.update(options)

        cli: list[str] = []
        cli_args: list[str] = []
        if verbose:
            cli.append("-" + "v" * verbose)
        for arg_name, value in args.items():
            option = self._known_options[arg_name]
            opt_cli = option["cli"]
            opt_type = option["type"]
            if opt_type == "boolean":
                if value:
                    cli.append(opt_cli)
            elif opt_type == "string":
                assert isinstance(value, str)
                cli.append(opt_cli + " %s")
                cli_args.append(value)
            elif opt_type == "json":
                cli.append(opt_cli + " %s")
                value_json = json.dumps(value)
                cli_args.append(value_json)
            else:
                raise TypeError(f"Unsupported argument type '{opt_type}'.")
        return " ".join(cli), cli_args

    def run_module(
        self,
        host: str,
        module_name: str,
        module_args: Optional[str],
        get_encoding: Optional[Callable[[], str]] = None,
        **options: Any,
    ) -> Any:
        cmd, args = "ansible --tree %s", []
        if self.inventory_file:
            cmd += " -i %s"
            args += [self.inventory_file]
        cmd += " -m %s"
        args += [module_name]
        if module_args:
            cmd += " --args %s"
            args += [module_args]
        options_cli, options_args = self.options_to_cli(options)
        if options_cli:
            cmd += " " + options_cli
            args.extend(options_args)
        cmd += " %s"
        args += [host]
        with tempfile.TemporaryDirectory() as d:
            args.insert(0, d)
            out = local.run_expect([0, 2, 8], cmd, *args)
            files = os.listdir(d)
            if not files and "skipped" in out.stdout.lower():
                return {
                    "failed": True,
                    "skipped": True,
                    "msg": "Skipped. You might want to try check=False",
                }
            if not files:
                raise RuntimeError(f"{out}")
            fpath = os.path.join(d, files[0])
            try:
                with open(fpath, encoding="ascii") as f:
                    return json.load(f)
            except UnicodeDecodeError:
                if get_encoding is None:
                    raise
                with open(fpath, encoding=get_encoding()) as f:
                    return json.load(f)

    @classmethod
    def get_runner(cls, inventory: Optional[str]) -> "AnsibleRunner":
        try:
            return cls._runners[inventory]
        except KeyError:
            cls._runners[inventory] = cls(inventory)
            return cls._runners[inventory]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from testinfra.backend import base


class LocalBackend(base.BaseBackend):
    NAME = "local"

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__("local", **kwargs)

    def get_pytest_id(self) -> str:
        return "local"

    @classmethod
    def get_hosts(cls, host: str, **kwargs: Any) -> list[str]:
        return [host]

    def run(self, command: str, *args: str, **kwargs: Any) -> base.CommandResult:
        return self.run_local(self.get_command(command, *args))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import salt.client
except ImportError:
    raise RuntimeError(
        "You must install salt package to use the salt backend"
    ) from None

from typing import Any, Optional

from testinfra.backend import base


class SaltBackend(base.BaseBackend):
    HAS_RUN_SALT = True
    NAME = "salt"

    def __init__(self, host: str, *args: Any, **kwargs: Any):
        self.host = host
        self._client: Optional[salt.client.LocalClient] = None
        super().__init__(self.host, *args, **kwargs)

    @property
    def client(self) -> salt.client.LocalClient:
        if self._client is None:
            self._client = salt.client.LocalClient()
        return self._client

    def run(self, command: str, *args: str, **kwargs: Any) -> base.CommandResult:
        command = self.get_command(command, *args)
        out = self.run_salt("cmd.run_all", [command])
        return self.result(
            out["retcode"],
            self.encode(command),
            stdout=out["stdout"],
            stderr=out["stderr"],
        )

    def run_salt(self, func: str, args: Any = None) -> Any:
        out = self.client.cmd(self.host, func, args or [])
        if self.host not in out:
            raise RuntimeError(
                f"Error while running {func}({args}): {out}. Minion not connected ?"
            )
        return out[self.host]

    @classmethod
    def get_hosts(cls, host: str, **kwargs: Any) -> list[str]:
        if host is None:
            host = "*"
        if any(c in host for c in "@*[?"):
            client = salt.client.LocalClient()
            if "@" in host:
                hosts = client.cmd(host, "test.true", tgt_type="compound").keys()
            else:
                hosts = client.cmd(host, "test.true").keys()
            if not hosts:
                raise RuntimeError(f"No host matching '{host}'")
            return sorted(hosts)
        return super().get_hosts(host, **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testinfra.modules.base import Module


class MountPoint(Module):
    """Test Mount Points"""

    def __init__(self, path, _attrs_cache=None):
        self.path = path
        self._attrs_cache = _attrs_cache
        super().__init__()

    @classmethod
    def _iter_mountpoints(cls):
        raise NotImplementedError

    @property
    def exists(self):
        """Return True if the mountpoint exists

        >>> host.mount_point("/").exists
        True

        >>> host.mount_point("/not/a/mountpoint").exists
        False

        """
        return bool(self._attrs)

    @property
    def _attrs(self):
        if self._attrs_cache is None:
            for mountpoint in self._iter_mountpoints():
                if mountpoint["path"] == self.path:
                    self._attrs_cache = mountpoint
                    break
            else:
                self._attrs_cache = {}
        return self._attrs_cache

    @property
    def filesystem(self):
        """Returns the filesystem type associated

        >>> host.mount_point("/").filesystem
        'ext4'

        """
        return self._attrs["filesystem"]

    @property
    def device(self):
        """Return the device associated

        >>> host.mount_point("/").device
        '/dev/sda1'

        """
        return self._attrs["device"]

    @property
    def options(self):
        """Return a list of options that a mount point has been created with

        >>> host.mount_point("/").options
        ['rw', 'relatime', 'data=ordered']

        """
        return self._attrs["options"]

    @classmethod
    def get_mountpoints(cls):
        """Returns a list of MountPoint instances

        >>> host.mount_point.get_mountpoints()
        [<MountPoint(path=/proc, device=proc, filesystem=proc, options=rw,nosuid,nodev,noexec,relatime)>,
         <MountPoint(path=/, device=/dev/sda1, filesystem=ext4, options=rw,relatime,errors=remount-ro,data=ordered)>]
        """
        mountpoints = []
        for mountpoint in cls._iter_mountpoints():
            mountpoints.append(cls(mountpoint["path"], mountpoint))
        return mountpoints

    @classmethod
    def get_module_class(cls, host):
        if host.system_info.type == "linux":
            return LinuxMountPoint
        if host.system_info.type.endswith("bsd"):
            return BSDMountPoint
        raise NotImplementedError

    def __repr__(self):
        if self.exists:
            d = self.device
            f = self.filesystem
            o = ",".join(self.options)
        else:
            d = ""
            f = ""
            o = ""
        return (
            f'<MountPoint(path="{self.path}", exists={self.exists}, device="{d}", '
            f'filesystem="{f}", options="{o}") at 0x{id(self):08x}>'
        )


class LinuxMountPoint(MountPoint):
    @classmethod
    def _iter_mountpoints(cls):
        check_output = cls(None).check_output
        for line in check_output("cat /proc/mounts").splitlines():
            splitted = line.split()
            # ignore rootfs
            # https://www.kernel.org/doc/Documentation/filesystems/ramfs-rootfs-initramfs.txt
            # suggests that most OS mount the filesystem over it, leaving
            # rootfs would result in ambiguity when resolving a mountpoint.
            if splitted[0] == "rootfs":
                continue

            yield {
                "path": splitted[1],
                "device": splitted[0],
                "filesystem": splitted[2],
                "options": splitted[3].split(","),
            }


class BSDMountPoint(MountPoint):
    @classmethod
    def _iter_mountpoints(cls):
        check_output = cls(None).check_output
        for line in check_output("mount -p").splitlines():
            splitted = line.split()
            yield {
                "path": splitted[1],
                "device": splitted[0],
                "filesystem": splitted[2],
                "options": splitted[3].split(","),
            }



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

try:
    import paramiko
except ImportError:
    raise RuntimeError(
        "You must install paramiko package (pip install paramiko) "
        "to use the paramiko backend"
    ) from None

import functools
from typing import Any, Optional

import paramiko.pkey
import paramiko.ssh_exception

from testinfra.backend import base


class IgnorePolicy(paramiko.MissingHostKeyPolicy):
    """Policy for ignoring missing host key."""

    def missing_host_key(
        self, client: paramiko.SSHClient, hostname: str, key: paramiko.pkey.PKey
    ) -> None:
        pass


class ParamikoBackend(base.BaseBackend):
    NAME = "paramiko"

    def __init__(
        self,
        hostspec: str,
        ssh_config: Optional[str] = None,
        ssh_identity_file: Optional[str] = None,
        timeout: int = 10,
        *args: Any,
        **kwargs: Any,
    ):
        self.host = self.parse_hostspec(hostspec)
        self.ssh_config = ssh_config
        self.ssh_identity_file = ssh_identity_file
        self.get_pty = False
        self.timeout = int(timeout)
        super().__init__(self.host.name, *args, **kwargs)

    def _load_ssh_config(
        self,
        client: paramiko.SSHClient,
        cfg: dict[str, Any],
        ssh_config: paramiko.SSHConfig,
        ssh_config_dir: str = "~/.ssh",
    ) -> None:
        for key, value in ssh_config.lookup(self.host.name).items():
            if key == "hostname":
                cfg[key] = value
            elif key == "user":
                cfg["username"] = value
            elif key == "port":
                cfg[key] = int(value)
            elif key == "identityfile":
                cfg["key_filename"] = os.path.expanduser(value[0])
            elif key == "stricthostkeychecking" and value == "no":
                client.set_missing_host_key_policy(IgnorePolicy())
            elif key == "requesttty":
                self.get_pty = value in ("yes", "force")
            elif key == "gssapikeyexchange":
                cfg["gss_auth"] = value == "yes"
            elif key == "gssapiauthentication":
                cfg["gss_kex"] = value == "yes"
            elif key == "proxycommand":
                cfg["sock"] = paramiko.ProxyCommand(value)
            elif key == "include":
                new_config_path = os.path.join(
                    os.path.expanduser(ssh_config_dir), value
                )
                with open(new_config_path) as f:
                    new_ssh_config = paramiko.SSHConfig()
                    new_ssh_config.parse(f)
                    self._load_ssh_config(client, cfg, new_ssh_config, ssh_config_dir)

    @functools.cached_property
    def client(self) -> paramiko.SSHClient:
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.WarningPolicy())
        cfg = {
            "hostname": self.host.name,
            "port": int(self.host.port) if self.host.port else 22,
            "username": self.host.user,
            "timeout": self.timeout,
            "password": self.host.password,
        }
        if self.ssh_config:
            ssh_config_dir = os.path.dirname(self.ssh_config)

            with open(self.ssh_config) as f:
                ssh_config = paramiko.SSHConfig()
                ssh_config.parse(f)
                self._load_ssh_config(client, cfg, ssh_config, ssh_config_dir)
        else:
            # fallback reading ~/.ssh/config
            default_ssh_config = os.path.join(os.path.expanduser("~"), ".ssh", "config")
            ssh_config_dir = os.path.dirname(default_ssh_config)

            try:
                with open(default_ssh_config) as f:
                    ssh_config = paramiko.SSHConfig()
                    ssh_config.parse(f)
            except OSError:
                pass
            else:
                self._load_ssh_config(client, cfg, ssh_config, ssh_config_dir)

        if self.ssh_identity_file:
            cfg["key_filename"] = self.ssh_identity_file
        client.connect(**cfg)  # type: ignore[arg-type]
        return client

    def _exec_command(self, command: bytes) -> tuple[int, bytes, bytes]:
        transport = self.client.get_transport()
        assert transport is not None
        chan = transport.open_session()
        if self.get_pty:
            chan.get_pty()
        chan.exec_command(command)
        stdout = b"".join(chan.makefile("rb"))
        stderr = b"".join(chan.makefile_stderr("rb"))
        rc = chan.recv_exit_status()
        return rc, stdout, stderr

    def run(self, command: str, *args: str, **kwargs: Any) -> base.CommandResult:
        command = self.get_command(command, *args)
        cmd = self.encode(command)
        try:
            rc, stdout, stderr = self._exec_command(cmd)
        except (paramiko.ssh_exception.SSHException, ConnectionResetError):
            transport = self.client.get_transport()
            assert transport is not None
            if not transport.is_active():
                # try to reinit connection (once)
                del self.client
                rc, stdout, stderr = self._exec_command(cmd)
            else:
                raise

        return self.result(rc, cmd, stdout, stderr)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from testinfra.backend import base


class PodmanBackend(base.BaseBackend):
    NAME = "podman"

    def __init__(self, name: str, *args: Any, **kwargs: Any):
        self.name, self.user = self.parse_containerspec(name)
        super().__init__(self.name, *args, **kwargs)

    def run(self, command: str, *args: str, **kwargs: Any) -> base.CommandResult:
        cmd = self.get_command(command, *args)
        if self.user is not None:
            out = self.run_local(
                "podman exec -u %s %s /bin/sh -c %s", self.user, self.name, cmd
            )
        else:
            out = self.run_local("podman exec %s /bin/sh -c %s", self.name, cmd)
        out.command = self.encode(cmd)
        return out



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re

from testinfra.modules.base import Module


def _re_match(line, regexp):
    match = regexp.match(line)
    if match is None:
        raise RuntimeError(f"could not parse {line}")
    return match.groups()


class Pip(Module):
    """Test pip package manager and packages"""

    def __init__(self, name, pip_path="pip"):
        self.name = name
        self.pip_path = pip_path
        super().__init__()

    @property
    def is_installed(self):
        """Test if the package is installed

        >>> host.package("pip").is_installed
        True
        """
        return self.run_test("%s show %s", self.pip_path, self.name).rc == 0

    @property
    def version(self):
        """Return package version as returned by pip

        >>> host.package("pip").version
        '18.1'
        """
        return self.check_output(
            "%s show %s | grep Version: | cut -d' ' -f2",
            self.pip_path,
            self.name,
        )

    @classmethod
    def check(cls, pip_path="pip"):
        """Verify installed packages have compatible dependencies.

        >>> cmd = host.pip.check()
        >>> cmd.rc
        0
        >>> cmd.stdout
        No broken requirements found.

        Can only be used if `pip check`_ command is available,
        for pip versions >= 9.0.0_.

        .. _pip check: https://pip.pypa.io/en/stable/reference/pip_check/
        .. _9.0.0: https://pip.pypa.io/en/stable/news/#id526
        """
        return cls.run_expect([0, 1], "%s check", pip_path)

    @classmethod
    def get_packages(cls, pip_path="pip"):
        """Get all installed packages and versions returned by `pip list`:

        >>> host.pip.get_packages(pip_path='~/venv/website/bin/pip')
        {'Django': {'version': '1.10.2'},
         'mywebsite': {'version': '1.0a3', 'path': '/srv/website'},
         'psycopg2': {'version': '2.6.2'}}
        """
        out = cls.run_expect([0, 2], "%s list --no-index --format=json", pip_path)
        pkgs = {}
        if out.rc == 0:
            for pkg in json.loads(out.stdout):
                # XXX: --output=json does not return install path
                pkgs[pkg["name"]] = {"version": pkg["version"]}
        else:
            # pip < 9
            output_re = re.compile(r"^(.+) \((.+)\)$")
            for line in cls.check_output("%s list --no-index", pip_path).splitlines():
                if line.startswith("Warning: "):
                    # Warning: cannot find svn location for ...
                    continue
                name, version = _re_match(line, output_re)
                if "," in version:
                    version, path = version.split(",", 1)
                    pkgs[name] = {"version": version, "path": path.strip()}
                else:
                    pkgs[name] = {"version": version}
        return pkgs

    @classmethod
    def get_outdated_packages(cls, pip_path="pip"):
        """Get all outdated packages with current and latest version

        >>> host.pip.get_outdated_packages(
        ...     pip_path='~/venv/website/bin/pip')
        {'Django': {'current': '1.10.2', 'latest': '1.10.3'}}
        """
        out = cls.run_expect([0, 2], "%s list -o --format=json", pip_path)
        pkgs = {}
        if out.rc == 0:
            for pkg in json.loads(out.stdout):
                pkgs[pkg["name"]] = {
                    "current": pkg["version"],
                    "latest": pkg["latest_version"],
                }
        else:
            # pip < 9
            # pip 8: pytest (3.4.2) - Latest: 3.5.0 [wheel]
            # pip < 8: pytest (Current: 3.4.2 Latest: 3.5.0 [wheel])
            regexpes = [
                re.compile(r"^(.+?) \((.+)\) - Latest: (.+) .*$"),
                re.compile(r"^(.+?) \(Current: (.+) Latest: (.+) .*$"),
            ]
            for line in cls.check_output("%s list -o", pip_path).splitlines():
                if line.startswith("Warning: "):
                    # Warning: cannot find svn location for ...
                    continue
                output_re = regexpes[1] if "Current:" in line else regexpes[0]
                name, current, latest = _re_match(line, output_re)
                pkgs[name] = {"current": current, "latest": latest}
        return pkgs



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testinfra.modules.base import InstanceModule


class Environment(InstanceModule):
    """Get Environment variables

    Example:

    >>> host.environment()
    {
        "EDITOR": "vim",
        "SHELL": "/bin/bash",
        [...]
    }
    """

    def __call__(self):
        ret_val = dict(
            i.split("=", 1) for i in self.check_output("env -0").split("\x00") if i
        )
        return ret_val

    def __repr__(self):
        return "<environment>"
